# coding: utf-8
# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref test for views."""

import json
import os
from datetime import date
from tempfile import NamedTemporaryFile
try:
    import unittest2 as unittest
except ImportError:
    import unittest

from yams.schema import role_name

from cubicweb import Binary
from cubicweb.devtools.testlib import CubicWebTC
from cubicweb.web.views.actions import CopyAction

from cubes.skos.rdfio import default_graph

from cubes.saem_ref.views.widgets import process_incomplete_date

import testutils


class SetDateTC(unittest.TestCase):

    def test_set_date_month_beginning(self):
        self.assertEqual(date(1997, 6, 5), process_incomplete_date("5/6/1997"))
        self.assertEqual(date(1997, 6, 5), process_incomplete_date("5/6/97"))
        self.assertEqual(date(2012, 6, 5), process_incomplete_date("5/6/12"))
        self.assertEqual(date(1997, 6, 1), process_incomplete_date("6/1997"))
        self.assertEqual(date(1997, 1, 1), process_incomplete_date("1997"))
        # XXX: behavior has changed in dateutil with dayfirst=True
        # https://github.com/dateutil/dateutil/commit/2d42e046d55b9fbbc0a2f41ce83fb8ec5de2d28b
        self.assertIn(process_incomplete_date("1997/6/5"), [date(1997, 6, 5),  # dateutil<=2.5.1
                                                            date(1997, 5, 6),  # dateutil>=2.5.2
                                                            ])

    def test_set_date_month_end(self):
        self.assertEqual(date(1994, 8, 28), process_incomplete_date("28/08/1994", True))
        self.assertEqual(date(1994, 8, 31), process_incomplete_date("08/1994", True))
        self.assertEqual(date(1994, 2, 28), process_incomplete_date("1994/02", True))
        self.assertEqual(date(1994, 12, 31), process_incomplete_date("1994", True))

    def test_set_date_failure(self):
        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("31/02/2012")
        self.assertIn("day is out of range for month", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("20/14/2015")
        self.assertIn("month must be in 1..12", str(cm.exception))

        with self.assertRaises(ValueError) as cm:
            process_incomplete_date("20")


class FuncViewsTC(CubicWebTC):

    def test_xmlwrap_component(self):
        with self.admin_access.cnx() as cnx:
            bob = testutils.authority_record(cnx, u'bob')
            uri = cnx.create_entity('ExternalUri', uri=u'http://logilab.fr')
            cnx.create_entity('EACResourceRelation',
                              resource_relation_agent=bob,
                              resource_relation_resource=uri,
                              xml_wrap=Binary('<plip>plop</plip>'))
            cnx.commit()
        with self.admin_access.web_request() as req:
            rset = req.find('EACResourceRelation')
            component = self.vreg['ctxcomponents'].select(
                'saem_ref.xml_wrap', req, rset=rset)
            content = []
            component.render_body(content.append)
        self.assertIn(u'plop', content[0])
        self.assertEqual(content[0].count(u'plip'), 2)

    def test_citation_view(self):
        with self.admin_access.cnx() as cnx:
            bob = testutils.authority_record(cnx, u'bob')
            mandate = cnx.create_entity('Mandate', term=u'u',
                                        mandate_agent=bob)
            citation = cnx.create_entity('Citation',
                                         note=u'plop',
                                         reverse_has_citation=mandate)
            citation_uri = cnx.create_entity('Citation',
                                             note=u'plop',
                                             uri=u'http://pl.op/',
                                             reverse_has_citation=mandate)
            cnx.commit()
            without_link_html = citation.view('citation-link')
            with_link_html = citation_uri.view('citation-link')
            self.assertEqual(
                with_link_html,
                u'<a class="truncate" href="http://pl.op/" title="plop">plop</a>')
            self.assertEqual(without_link_html, u'<i class="truncate">plop</i>')

    def test_eac_import_user_has_no_naa(self):
        regid = 'saem_ref.eac-import'
        with self.admin_access.web_request() as req:
            output = self.view(regid, req=req, template=None)
            self.assertIn('be in an organization', output)
            self.assertIn('to access this functionnality', output)
        with self.admin_access.web_request() as req:
            authority = req.cnx.find('Organization', name=u'Default authority').one()
            req.user.cw_set(authority=authority)
            req.cnx.commit()
            output = self.view(regid, req=req, template=None)
            self.assertIn('have an NAA configured', output)
            self.assertIn('to access this functionnality', output)

    def test_eac_import_ok(self):
        regid = 'saem_ref.eac-import'
        fname = 'FRAD033_EAC_00001_simplified.xml'
        with self.admin_access.web_request() as req:
            req.user.cw_set(authority=testutils.authority_with_naa(req.cnx))
            naa = req.cnx.create_entity('ArkNameAssigningAuthority', who=u'EAC', what=1)
            # simply test the form properly render and is well formed
            self.view(regid, req=req, template=None)
            fields = {'file': (fname, open(self.datapath('EAC/' + fname))),
                      'naa': unicode(naa.eid)}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            req.view(regid)
            # test AuthorityRecord and Agent have been created, the former with a ARK from NAA
            # specified in form params, the Agent with a ARK from the user's NAA
            arecord = req.find('AuthorityRecord', name=u'Gironde. Conseil général').one()
            self.assertTrue(arecord.ark.startswith('1/'), arecord.ark)
            agent = req.find('Agent', name=u'Delphine Jamet').one()
            self.assertTrue(agent.ark.startswith('0/'), agent.ark)

    def test_eac_non_unique_isni(self):
        regid = 'saem_ref.eac-import'
        fname = 'FRAD033_EAC_00001_simplified.xml'
        with self.admin_access.client_cnx() as cnx:
            # ISNI is the same as the agent in EAC file.
            testutils.authority_record(cnx, u'bob', isni=u'22330001300016')
            cnx.commit()
            rset = cnx.find('AuthorityRecord')
            self.assertEqual(len(rset), 1)
        with self.admin_access.web_request() as req:
            req.user.cw_set(authority=testutils.authority_with_naa(req.cnx))
            # simply test the form properly render and is well formed
            self.view(regid, req=req, template=None)
            fields = {'file': (fname, open(self.datapath('EAC/' + fname))),
                      'naa': unicode(req.user.naa.eid)}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn('EAC import failed', html)
            # Still only one AuthorityRecord.
            rset = req.find('AuthorityRecord')
            self.assertEqual(len(rset), 1)

    def test_eac_invalid_xml(self):
        regid = 'saem_ref.eac-import'
        fname = 'invalid_xml.xml'
        with self.admin_access.web_request() as req:
            req.user.cw_set(authority=testutils.authority_with_naa(req.cnx))
            fields = {'file': (fname, open(self.datapath('EAC/' + fname))),
                      'naa': unicode(req.user.naa.eid)}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn('Invalid XML file', html)

    def test_eac_missing_tag(self):
        regid = 'saem_ref.eac-import'
        fname = 'missing_tag.xml'
        with self.admin_access.web_request() as req:
            req.user.cw_set(authority=testutils.authority_with_naa(req.cnx))
            fields = {'file': (fname, open(self.datapath('EAC/' + fname))),
                      'naa': unicode(req.user.naa.eid)}
            req.form = self.fake_form(regid, fields)
            # now actually test the import
            html = req.view(regid)
            self.assertIn('Missing tag cpfDescription in XML file', html)

    def test_highlight_script_execution(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = testutils.setup_scheme(cnx, u'my thesaurus')
            cnx.commit()
            url = scheme.absolute_url(highlight='toto')
        self.assertIn(
            '$(document).ready(function(){$("h1, h2, h3, h4, h5, table tbody td")'
            '.highlight("toto");});}',
            self.http_publish(url)[0])

    def test_highlight_on_rql_plain_text_search_same_etype(self):
        with self.admin_access.client_cnx() as cnx:
            # we need two concepts to get a list view
            scheme = testutils.setup_scheme(cnx, u'my thesaurus', u'toto', u'tata toto')
            concept = scheme.reverse_in_scheme[0]
            cnx.commit()
        url = 'http://testing.fr/cubicweb/view?rql=toto&__fromsearchbox=1&subvid=tsearch'
        html = self.http_publish(url)[0]
        self.assertIn(
            '<a href="http://testing.fr/cubicweb/%s?highlight=toto" title="">' % concept.eid,
            html)

    def test_skos_negociation(self):
        with self.admin_access.client_cnx() as cnx:
            scheme = cnx.create_entity('ConceptScheme', title=u'musique',
                                       ark_naa=testutils.naa(cnx))
            scheme.add_concept(u'pop')
            cnx.commit()
        with self.admin_access.web_request(headers={'Accept': 'application/rdf+xml'}) as req:
            result = self.app_handle_request(req, 'conceptscheme')
            with NamedTemporaryFile(delete=False) as fobj:
                try:
                    fobj.write(result)
                    fobj.close()
                    graph = default_graph()
                    graph.load('file://' + fobj.name, rdf_format='xml')
                finally:
                    os.unlink(fobj.name)

    def test_download_filename(self):
        with self.admin_access.web_request() as req:
            cnx = req.cnx
            agent = testutils.authority_record(cnx, u'jim')
            for ark, expected_filename in (
                (u"", "EAC_{0}.xml".format(agent.eid)),
                (u"ZZZ/4242", "EAC_ZZZ_4242.xml".format(agent.eid)),
            ):
                agent.cw_set(ark=ark)
                view = self.vreg['views'].select('saem_ref.eac', req, agent.as_rset())
                view.set_request_content_type()
                self.assertEqual(
                    view._cw.headers_out.getRawHeaders('content-disposition'),
                    ['attachment;filename="{0}"'.format(expected_filename)],
                )

    def test_agent_place_as_concept_view(self):
        with self.admin_access.web_request() as req:
            cnx = req.cnx
            agent_place = cnx.create_entity('AgentPlace', name=u"""çàè€É'"><$""")
            content = agent_place.view('saem.agent_place_as_concept')
            self.assertEqual(content,
                             u"<strong><span>çàè€É&#39;&quot;&gt;&lt;$</span></strong>")

    def test_listitem(self):
        with self.admin_access.web_request() as req:
            entity = req.create_entity('CWGroup', name=u'a')
            content = entity.view('saem.listitem', tabid='blah')
            self.assertIn(entity.view('oneline'), content)
            self.assertIn('__redirectparams=tab%3Dblah&amp', content)


class AuthorityRecordFormsTC(CubicWebTC):

    def test_no_copy_action(self):
        with self.admin_access.cnx() as cnx:
            testutils.authority_record(cnx, u'toto')
            cnx.commit()
        with self.admin_access.web_request() as req:
            rset = req.find('AuthorityRecord', name=u'toto')
            actions = self.pactionsdict(req, rset)
            self.assertNotIn(CopyAction, actions['moreactions'])

    def test_unrelated_authorityrecord(self):
        from cubes.saem_ref.views.authorityrecord import unrelated_authorityrecord
        with self.admin_access.cnx() as cnx:
            ar1 = testutils.authority_record(cnx, u'1').eid
            ar2 = testutils.authority_record(cnx, u'2').eid
            # ExternalUri should not appear in choices.
            cnx.create_entity('ExternalUri', uri=u'http://no-where')
            cnx.commit()
        with self.admin_access.web_request() as req:
            cls = self.vreg['etypes'].etype_class('HierarchicalRelation')(req)
            form = self.vreg['forms'].select('edition', req, entity=cls)
            field = form.field_by_name('hierarchical_parent', 'subject')
            choices = unrelated_authorityrecord(
                'hierarchical_parent', form, field)
            expected = [(u'1', unicode(ar1)), (u'2', unicode(ar2))]
            self.assertCountEqual(choices, expected)
        # Now with a __linkto.
        linkto = 'hierarchical_parent:{0}:subject'.format(ar2)
        with self.admin_access.web_request(__linkto=linkto) as req:
            cls = self.vreg['etypes'].etype_class('HierarchicalRelation')(req)
            form = self.vreg['forms'].select('edition', req, entity=cls)
            field = form.field_by_name('hierarchical_parent', 'subject')
            choices = unrelated_authorityrecord(
                'hierarchical_parent', form, field)
            expected = [(u'2', unicode(ar2))]
            self.assertCountEqual(choices, expected)
            field = form.field_by_name('hierarchical_child', 'subject')
            choices = unrelated_authorityrecord(
                'hierarchical_child', form, field)
            expected = [(u'1', unicode(ar1)), (u'2', unicode(ar2))]
            self.assertCountEqual(choices, expected)


class SEDANavigationTC(CubicWebTC):

    def test_breadcrumbs(self):
        with self.admin_access.cnx() as cnx:
            scheme = testutils.seda_scheme(
                cnx, u'seda_document_type_code', u'preferred', u'QWE')
            doctypecodevalue = scheme.reverse_in_scheme[0]
            doctypecode = cnx.create_entity(
                'SEDADocumentTypeCode',
                seda_document_type_code_value=doctypecodevalue)
            doc = cnx.create_entity('ProfileDocument',
                                    seda_document_type_code=doctypecode)
            profile = cnx.create_entity('SEDAProfile', ark_naa=testutils.naa(cnx))
            archobj = cnx.create_entity(
                'ProfileArchiveObject',
                seda_name=cnx.create_entity('SEDAName'),
                seda_parent=profile)
            cnx.commit()
        with self.admin_access.web_request() as req:
            archobj = req.entity_from_eid(archobj.eid)
            profile = req.entity_from_eid(profile.eid)
            doc = req.entity_from_eid(doc.eid)
            # ProfileDocument not related through seda_parent, breadcrumbs
            # leads to /sedalib.
            breadcrumbs = doc.cw_adapt_to('IBreadCrumbs').breadcrumbs()
            expected_breadcrumbs = [
                (u'http://testing.fr/cubicweb/sedalib', u'SEDA components'),
                doc,
            ]
            self.assertEqual(breadcrumbs, expected_breadcrumbs)
            # ProfileDocument has seda_parent, breadcrumbs leads to parents
            # (ProfileArchiveObject, SEDAProfile).
            doc.cw_set(seda_parent=archobj)
            req.cnx.commit()
            doc.cw_clear_all_caches()
            breadcrumbs = doc.cw_adapt_to('IBreadCrumbs').breadcrumbs()
            expected_breadcrumbs = [profile, archobj, doc]
            self.assertEqual(breadcrumbs, expected_breadcrumbs)


class SEDAFormsTC(CubicWebTC):

    def _build_seda_scheme(self, rtype_name, label_kind=u'preferred'):
        with self.admin_access.cnx() as cnx:
            labels = (u'pif', u'paf', u'pouf')
            testutils.seda_scheme(cnx, rtype_name, label_kind, *labels)
            cnx.commit()
        return labels

    def test_seda_document_type_code_vocabulary(self):
        """Check that we get correct values in the combo box for document type code."""
        labels = self._build_seda_scheme(u'seda_document_type_code')
        with self.admin_access.web_request() as req:
            doc = self.vreg['etypes'].etype_class('SEDADocumentTypeCode')(req)
            form = self.vreg['forms'].select('edition', req, entity=doc)
            field = form.field_by_name('seda_document_type_code_value', 'subject')
            form_labels = [label for label, _ in field.vocabulary(form)]
            self.assertCountEqual(labels, form_labels)

    def test_seda_file_type_code_vocabulary(self):
        """Check that we get correct values in the combo box for file type code."""
        labels = self._build_seda_scheme(u'seda_file_type_code',
                                         u'alternative')
        with self.admin_access.web_request() as req:
            doc = self.vreg['etypes'].etype_class('SEDAFileTypeCode')(req)
            form = self.vreg['forms'].select('edition', req, entity=doc)
            field = form.field_by_name('seda_file_type_code_value', 'subject')
            values = list(field.vocabulary(form))
            form_labels = [label for label, _ in values]
            self.assertCountEqual(form_labels, labels)

    def test_seda_character_set_code_vocabulary(self):
        """Check that we get correct values in the combo box for character set code."""
        labels = self._build_seda_scheme(u'seda_character_set_code',
                                         u'alternative')
        with self.admin_access.web_request() as req:
            doc = self.vreg['etypes'].etype_class('SEDACharacterSetCode')(req)
            form = self.vreg['forms'].select('edition', req, entity=doc)
            field = form.field_by_name('seda_character_set_code_value', 'subject')
            values = list(field.vocabulary(form))
            form_labels = [label for label, _ in values]
            self.assertCountEqual(form_labels, labels)

    def test_seda_get_related_version(self):
        """Check that we get correct results when asking for `draft`, `published`, `replaced`
        version of a profile."""
        with self.admin_access.web_request() as req:
            profile1 = testutils.publishable_profile(req, title=u'Profile 1')
            req.cnx.commit()
            profile1.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile1.cw_clear_all_caches()
            profile2 = testutils.publishable_profile(req, title=u'Profile 2', seda_replace=profile1)
            req.cnx.commit()
            profile2.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile2.cw_clear_all_caches()
            profile3 = testutils.publishable_profile(req, title=u'Profile 3', seda_replace=profile2)
            req.cnx.commit()
            profile3.cw_adapt_to('IWorkflowable').fire_transition('publish')
            req.cnx.commit()
            profile3.cw_clear_all_caches()
            profile4 = testutils.publishable_profile(req, title=u'Profile 4', seda_replace=profile3)
            req.cnx.commit()

            def unwrap_generator(gen):
                try:
                    return next(iter(gen))
                except StopIteration:
                    return None

            # Draft profile
            box4 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile4)
            self.assertEqual(unwrap_generator(box4.predecessor()).eid, profile3.eid)
            self.assertIsNone(unwrap_generator(box4.current_version(state=u'published')))
            self.assertIsNone(unwrap_generator(box4.current_version(state=u'draft')))
            # Published profile
            box3 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile3)
            self.assertEqual(unwrap_generator(box3.predecessor()).eid, profile2.eid)
            self.assertIsNone(unwrap_generator(box3.current_version(state=u'published')))
            self.assertEqual(unwrap_generator(box3.current_version(state=u'draft')).eid,
                             profile4.eid)
            # Deprecated profile
            box2 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile2)
            self.assertIsNone(unwrap_generator(box2.predecessor()))
            self.assertEqual(unwrap_generator(box2.current_version(state=u'published')).eid,
                             profile3.eid)
            self.assertEqual(unwrap_generator(box2.current_version(state=u'draft')).eid,
                             profile4.eid)
            # Older deprecated profile
            box1 = self.vreg['ctxcomponents'].select('saem.seda.relatedprofiles', req,
                                                     entity=profile1)
            self.assertIsNone(unwrap_generator(box1.predecessor()))
            self.assertEqual(unwrap_generator(box1.current_version(state=u'published')).eid,
                             profile3.eid)
            self.assertEqual(unwrap_generator(box1.current_version(state=u'draft')).eid,
                             profile4.eid)

    def test_seda_actions_notype_to_import(self):
        from cubes.saem_ref.views import seda

        with self.admin_access.web_request() as req:
            profile = testutils.setup_seda_profile(req)
            req.cnx.commit()
            profile._cw = req  # XXX
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertIn(seda.SEDAImportProfileArchiveObjectAction, actions['moreactions'])
            self.assertNotIn(seda.SEDAImportProfileDocumentAction, actions['moreactions'])
            actions = self.pactionsdict(req, profile.archives[0].as_rset())
            self.assertIn(seda.SEDAImportProfileArchiveObjectAction, actions['moreactions'])
            self.assertIn(seda.SEDAImportProfileDocumentAction, actions['moreactions'])
        with self.new_access('anon').web_request() as req:
            profile = req.entity_from_eid(profile.eid)
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertNotIn(seda.SEDAImportProfileArchiveObjectAction, actions['moreactions'])
            self.assertNotIn(seda.SEDAImportProfileDocumentAction, actions['moreactions'])

    def test_seda_actions_not_exportable(self):
        from cubes.saem_ref.views import seda

        with self.admin_access.web_request() as req:
            profile = req.create_entity('SEDAProfile', ark_naa=testutils.naa(req.cnx))
            req.cnx.commit()
            profile._cw = req  # XXX
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertNotIn(seda.SEDA1DownloadAction, actions['moreactions'])
            self.assertNotIn(seda.SEDA02DownloadAction, actions['moreactions'])
            req.create_entity('ProfileArchiveObject',
                              seda_name=req.create_entity('SEDAName'),
                              seda_parent=profile)
            req.cnx.commit()
            profile.cw_clear_all_caches()
        with self.admin_access.web_request() as req:
            profile = req.entity_from_eid(profile.eid)
            actions = self.pactionsdict(req, profile.as_rset())
            self.assertIn(seda.SEDA1DownloadAction, actions['moreactions'])
            self.assertIn(seda.SEDA02DownloadAction, actions['moreactions'])


class SEDAComponentsImportTC(CubicWebTC):

    def test_seda_doimport_controller(self):
        """Test for 'saem_ref.seda.doimport' controller (called from JavaScript).
        """
        with self.admin_access.cnx() as cnx:
            scheme = testutils.seda_scheme(
                cnx, u'seda_document_type_code', u'preferred', u'QWE')
            doctypecodevalue = scheme.reverse_in_scheme[0]
            doctypecode = cnx.create_entity(
                'SEDADocumentTypeCode',
                seda_document_type_code_value=doctypecodevalue)
            doc = cnx.create_entity('ProfileDocument',
                                    user_cardinality=u'1',
                                    user_annotation=u'plop',
                                    seda_document_type_code=doctypecode)
            profile = cnx.create_entity('SEDAProfile', ark_naa=testutils.naa(cnx))
            archobj = cnx.create_entity(
                'ProfileArchiveObject',
                seda_name=cnx.create_entity('SEDAName'),
                seda_parent=profile)
            cnx.commit()
        params = dict(eid=archobj.eid, cloned=doc.eid)
        with self.admin_access.web_request(**params) as req:
            path, _ = self.expect_redirect_handle_request(
                req, 'saem_ref.seda.doimport')
            etype, eid = path.split('/')
            self.assertEqual(etype, 'ProfileArchiveObject'.lower())
            clone = req.execute('Any X WHERE X seda_parent P, P eid %(p)s',
                                {'p': eid}).one()
            self.assertEqual([x.eid for x in clone.seda_clone_of], [doc.eid])
            # Check that relationship targets have been copied.
            self.assertEqual(
                clone.seda_document_type_code[0].seda_document_type_code_value[0].eid,
                doctypecodevalue.eid)
            # Check that original entity attributes have been copied.
            self.assertEqual(clone.user_cardinality, u'1')
            self.assertEqual(clone.user_annotation, u'plop')


class ArkViewsTC(CubicWebTC):

    def test_ark_authority_record_creation(self):
        with self.admin_access.cnx() as cnx:
            org = testutils.authority_with_naa(cnx)
            naa_eid = org.ark_naa[0].eid
            cnx.commit()
        with self.admin_access.web_request() as req:
            akind = req.cnx.find('AgentKind', name=u'person').one()
            record = self.vreg['etypes'].etype_class('AuthorityRecord')(req)
            record.eid = 'A'
            fields = {
                role_name('name', 'subject'): u'007',
                role_name('ark', 'subject'): u'',
                role_name('agent_kind', 'subject'): str(akind.eid),
                role_name('ark_naa', 'subject'): str(naa_eid),
            }
            req.form = self.fake_form('edition', entity_field_dicts=[(record, fields)])
            eid = int(self.expect_redirect_handle_request(req)[0])
            record = req.cnx.entity_from_eid(eid)
            self.assertEqual(record.ark, u'0/r%09d' % eid)

    def test_ark_agent_creation(self):
        with self.admin_access.web_request() as req:
            org = testutils.authority_with_naa(req)
            agent = self.vreg['etypes'].etype_class('Agent')(req)
            agent.eid = 'A'
            fields = {role_name('name', 'subject'): u'007',
                      role_name('ark', 'subject'): u'',
                      role_name('authority', 'subject'): str(org.eid)}
            req.form = self.fake_form('edition', entity_field_dicts=[(agent, fields)])
            eid = int(self.expect_redirect_handle_request(req)[0])
            agent = req.cnx.entity_from_eid(eid)
            self.assertEqual(agent.ark, u'0/oa%09d' % eid)

    def test_ark_scheme_creation(self):
        with self.admin_access.web_request() as req:
            scheme = self.vreg['etypes'].etype_class('ConceptScheme')(req)
            scheme.eid = 'A'
            fields = {role_name('ark', 'subject'): u'',
                      role_name('ark_naa', 'subject'): unicode(testutils.naa(req.cnx).eid)}
            req.form = self.fake_form('edition', entity_field_dicts=[(scheme, fields)])
            eid = int(self.expect_redirect_handle_request(req)[0])
            scheme = req.cnx.entity_from_eid(eid)
            self.assertEqual(scheme.ark, u'0/v%09d' % eid)

    def test_ark_concept_creation_form(self):
        # test addition of a concept by specifying in_scheme in form
        with self.admin_access.web_request() as req:
            scheme = req.cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(req.cnx))
            concept = self.vreg['etypes'].etype_class('Concept')(req)
            concept.eid = 'A'
            concept_fields = {role_name('in_scheme', 'subject'): str(scheme.eid),
                              role_name('ark', 'subject'): u''}
            label = self.vreg['etypes'].etype_class('Label')(req)
            label.eid = 'B'
            label_fields = {role_name('label', 'subject'): u'Hello',
                            role_name('label_of', 'subject'): 'A'}
            req.form = self.fake_form('edition', entity_field_dicts=[(concept, concept_fields),
                                                                     (label, label_fields)])
            eid = int(self.expect_redirect_handle_request(req)[0])
            concept = req.cnx.entity_from_eid(eid)
            self.assertEqual(concept.ark, u'0/c%09d' % eid)

    def test_ark_concept_creation_linkto(self):
        # test addition of a concept by specifying in_scheme with __linkto
        with self.admin_access.web_request() as req:
            scheme = req.cnx.create_entity('ConceptScheme', ark_naa=testutils.naa(req.cnx))
            concept = self.vreg['etypes'].etype_class('Concept')(req)
            concept.eid = 'A'
            concept_fields = {}
            label = self.vreg['etypes'].etype_class('Label')(req)
            label.eid = 'B'
            label_fields = {role_name('label', 'subject'): u'Goodby',
                            role_name('label_of', 'subject'): 'A'}
            req.form = self.fake_form('edition', entity_field_dicts=[(concept, concept_fields),
                                                                     (label, label_fields)])
            req.form['__linkto'] = 'in_scheme:%s:subject' % scheme.eid
            eid = int(self.expect_redirect_handle_request(req)[0])
            concept = req.cnx.entity_from_eid(eid)
            self.assertEqual(concept.ark, u'0/c%09d' % eid)

    def test_ark_url_rewrite(self):
        with self.admin_access.web_request() as req:
            rewriter = self.vreg['urlrewriting'].select('schemabased', req)
            _pmid, rset = rewriter.rewrite(req, u'/ark:/JoE/Dalton')
            self.assertEqual(rset.printable_rql(), 'Any X WHERE X ark "JoE/Dalton"')

    def test_ark_ws(self):
        with self.new_access('anon').web_request(headers={'Accept': 'application/json'},
                                                 method='POST') as req:
            # not authenticated
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': 'This service requires authentication.'}])
        with self.admin_access.web_request(headers={'Accept': 'application/json'}) as req:
            # authenticated but user has no authority
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': 'User is not associated to an authority.'}])
        with self.admin_access.web_request(headers={'Accept': 'application/json'}) as req:
            authority = testutils.authority_with_naa(req)
            req.user.cw_set(authority=authority)
            req.cnx.commit()
            # authenticated with authority, but noot using a POST
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'error': 'This service is only accessible using POST.'}])
        with self.admin_access.web_request(headers={'Accept': 'application/json'},
                                           method='POST') as req:
            # eventually
            result = self.app_handle_request(req, 'ark')
            self.assertEqual(json.loads(result),
                             [{'ark': '0/a000000001'}])

    def test_download_filename(self):
        with self.admin_access.web_request() as req:
            cnx = req.cnx
            record = testutils.authority_record(cnx, u'jim')
            for ark, expected_filename in (
                (u"", "EAC_{0}.xml".format(record.eid)),
                (u"ZZZ/4242", "EAC_ZZZ_4242.xml"),
            ):
                record.cw_set(ark=ark)
                view = self.vreg['views'].select('saem_ref.eac', req, record.as_rset())
                view.set_request_content_type()
                self.assertEqual(
                    view._cw.headers_out.getRawHeaders('content-disposition'),
                    ['attachment;filename="{0}"'.format(expected_filename)],
                )


if __name__ == '__main__':
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views for import/export of AuthorityRecord from/to EAC"""

import os.path

from cubicweb import tags
from cubicweb.view import View
from cubicweb.predicates import is_instance, match_user_groups, one_line_rset
from cubicweb.web import action, formfields as ff, formwidgets as fw, httpcache
from cubicweb.web.views import cwsources, forms, idownloadable

from cubes.skos import to_unicode
from cubicweb.dataimport.importer import HTMLImportLog

from cubes.saem_ref import user_has_naa
from cubes.saem_ref.dataimport import eac

_ = unicode


class EACImportMixin(object):
    __regid__ = 'saem_ref.eac-import'
    __select__ = match_user_groups('managers', 'users') & user_has_naa()


class EACImportViewNoNaa(View):
    __regid__ = 'saem_ref.eac-import'
    __select__ = match_user_groups('managers', 'users') & ~user_has_naa()

    def call(self):
        self.w(tags.h1(self._cw._('Importing an AuthorityRecord from a EAC-CPF file')))
        if not self._cw.user.authority:
            msg = self._cw._("You must <a href='{0}'>be in an organization</a> to access "
                             "this functionnality.")
            url = self._cw.user.absolute_url(vid='edition')
        else:
            msg = self._cw._("Your organization must <a href='{0}'>have an NAA configured</a> to "
                             "access this functionnality.")
            url = self._cw.user.authority[0].absolute_url(vid='edition')
        self.w(tags.div(msg.format(url)))


def naa_form_vocabulary(form, field):
    """Field vocabulary function returning the list of available authorities"""
    rset = form._cw.execute('Any XN, X ORDERBY XN WHERE X is ArkNameAssigningAuthority, X who XN')
    return [(name, unicode(eid)) for name, eid in rset]


class EACImportForm(EACImportMixin, forms.FieldsForm):
    """File import form for EAC-CPF"""
    filefield = ff.FileField(name='file', label=_('EAC-CPF file'),
                             required=True)
    naafield = ff.StringField(name='naa', required=True,
                              choices=naa_form_vocabulary, sort=False)
    form_buttons = [fw.SubmitButton(label=_('do_import'))]

    @property
    def action(self):
        return self._cw.build_url(vid=self.__regid__)


class EACImportView(EACImportMixin, View):
    """EAC-CPF import controller"""

    def call(self):
        self.w(tags.h1(self._cw._('Importing an AuthorityRecord from a EAC-CPF file')))
        form = self._cw.vreg['forms'].select(self.__regid__, self._cw)
        if form.posting:
            posted = form.process_posted()
            naa = self._cw.cnx.entity_from_eid(posted['naa'])
            stream = posted['file']
            import_log = HTMLImportLog(os.path.basename(stream.filename))
            try:
                _, _, eid = self._cw.cnx.call_service(self.__regid__, stream=stream,
                                                      naa=naa,
                                                      import_log=import_log)
            except eac.InvalidXML as exc:
                msg = self._cw._('Invalid XML file')
                self.exception('error while importing %s', stream.filename)
                mtype = 'danger'
                import_log.record_fatal(
                    self._cw._('xml syntax error: ') + to_unicode(exc))
            except eac.MissingTag as exc:
                if exc.tag_parent:
                    err = self._cw._('Missing tag %(tag)s within element %(parent)s in XML file')
                    params = {'tag': exc.tag, 'parent': exc.tag_parent}
                    msg = err % params
                else:
                    err = self._cw._('Missing tag %(tag)s in XML file')
                    params = {'tag': exc.tag}
                    msg = err % params
                self.exception('error while importing %s', stream.filename)
                mtype = 'danger'
                import_log.record_fatal(err % params)
            except Exception as exc:  # pylint: disable=broad-except
                msg = self._cw._('EAC import failed')
                self.exception('error while importing %s', stream.filename)
                mtype = 'danger'
            else:
                agent = self._cw.find('AuthorityRecord', eid=eid).one()
                msg = (self._cw._('EAC-CPF import completed: %s') %
                       agent.view('oneline'))
                mtype = 'success'
            self.w(tags.div(msg, klass="alert alert-%s" % mtype))
            if import_log.logs:
                self._cw.view('cw.log.table',
                              pyvalue=cwsources.log_to_table(
                                  self._cw, u''.join(import_log.logs)),
                              default_level='Warning', w=self.w)
        else:
            form.render(w=self.w)


# EAC export

class EACExportAction(action.Action):
    __regid__ = 'saem_ref.eac'
    __select__ = (action.Action.__select__
                  & one_line_rset()
                  & is_instance('AuthorityRecord'))

    title = _('EAC export')
    category = 'moreactions'
    seda_download_vid = 'saem_ref.eac'

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid=self.seda_download_vid)


class EACDownloadView(idownloadable.DownloadView):
    """EAC download view"""
    __regid__ = 'saem_ref.eac'
    __select__ = one_line_rset() & is_instance('AuthorityRecord')

    http_cache_manager = httpcache.NoHTTPCacheManager
    adapter_id = 'EAC-CPF'

    def set_request_content_type(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.adapter_id)
        self._cw.set_content_type(adapter.content_type, filename=adapter.file_name,
                                  encoding=adapter.encoding, disposition='attachment')

    def call(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.adapter_id)
        self.w(adapter.dump())



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for entities.container"""

from datetime import date

from cubicweb.devtools.testlib import CubicWebTC

from cubes.saem_ref.entities import container

import testutils


def sort_container(container_def):
    for k, v in container_def:
        yield k, sorted(v)


class ContainerTC(CubicWebTC):

    def test_authorityrecord_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = container.authority_record_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def,
                         {'Activity': [('generated', 'subject'), ('used', 'subject')],
                          'AgentFunction': [('function_agent', 'subject')],
                          'AgentPlace': [('place_agent', 'subject')],
                          'Citation': [('has_citation', 'object')],
                          'EACResourceRelation': [('resource_relation_agent', 'subject')],
                          'EACSource': [('source_agent', 'subject')],
                          'GeneralContext': [('general_context_of', 'subject')],
                          'History': [('history_agent', 'subject')],
                          'LegalStatus': [('legal_status_agent', 'subject')],
                          'Mandate': [('mandate_agent', 'subject')],
                          'Occupation': [('occupation_agent', 'subject')],
                          'PostalAddress': [('place_address', 'object')],
                          'Structure': [('structure_agent', 'subject')]})
        entity = self.vreg['etypes'].etype_class('AuthorityRecord')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        self.assertIsNone(entity.cw_adapt_to('IContained'))

    def test_scheme_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = container.scheme_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def,
                         {'Activity': [('generated', 'subject'), ('used', 'subject')],
                          'Concept': [('in_scheme', 'subject')]})
        entity = self.vreg['etypes'].etype_class('ConceptScheme')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        self.assertIsNone(entity.cw_adapt_to('IContained'))

    def test_concept_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = container.concept_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def,
                         {'Activity': [('generated', 'subject'), ('used', 'subject')],
                          'Label': [('label_of', 'subject')]})
        entity = self.vreg['etypes'].etype_class('Concept')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        # Concept is both container and contained :
        self.assertIsNotNone(entity.cw_adapt_to('IContained'))

    def test_seda_profile_container(self):
        # line below should be copied from entities.container.registration_callback
        container_def = container.seda_profile_container_def(self.schema)
        container_def = dict(sort_container(container_def))
        self.assertEqual(container_def,
                         {'SEDACharacterSetCode': [('seda_character_set_code', 'object')],
                          'SEDADocumentTypeCode': [('seda_document_type_code', 'object')],
                          'ProfileArchiveObject': [('seda_parent', 'subject')],
                          'ProfileDocument': [('seda_parent', 'subject')],
                          'SEDAAccessRestrictionCode': [('seda_access_restriction_code', 'object')],
                          'SEDAAppraisalRule': [('seda_appraisal_rule', 'object')],
                          'SEDAAppraisalRuleCode': [('seda_appraisal_rule_code', 'object')],
                          'SEDAAppraisalRuleDuration': [('seda_appraisal_rule_duration', 'object')],
                          'SEDAContentDescription': [('seda_content_description', 'object')],
                          'SEDADate': [('seda_latest_date', 'object'),
                                       ('seda_oldest_date', 'object')],
                          'SEDADescription': [('seda_description', 'object')],
                          'SEDADescriptionLevel': [('seda_description_level', 'object')],
                          'SEDAFileTypeCode': [('seda_file_type_code', 'object')],
                          'SEDAKeyword': [('seda_keyword_of', 'subject')],
                          'SEDAName': [('seda_name', 'object')],
                          })
        entity = self.vreg['etypes'].etype_class('SEDAProfile')(self)
        self.assertIsNotNone(entity.cw_adapt_to('IContainer'))
        self.assertIsNone(entity.cw_adapt_to('IContained'))


class TreeTC(CubicWebTC):

    maxDiff = None

    def test_authorityrecord_tree(self):
        tree_def = container.tree_def(self.schema, 'AuthorityRecord',
                                      skiprtypes=['wf_info_for']).items()
        tree_def = dict(sort_container(tree_def))
        self.assertEqual(tree_def,
                         {'Activity': [],
                          'AuthorityRecord': [('function_agent', 'object'),
                                              ('general_context_of', 'object'),
                                              ('generated', 'object'),
                                              ('history_agent', 'object'),
                                              ('legal_status_agent', 'object'),
                                              ('mandate_agent', 'object'),
                                              ('occupation_agent', 'object'),
                                              ('place_agent', 'object'),
                                              ('resource_relation_agent', 'object'),
                                              ('source_agent', 'object'),
                                              ('structure_agent', 'object'),
                                              ('used', 'object')],
                          'AgentFunction': [('has_citation', 'subject')],
                          'AgentPlace': [('has_citation', 'subject'), ('place_address', 'subject')],
                          'Citation': [],
                          'EACResourceRelation': [],
                          'EACSource': [],
                          'GeneralContext': [('has_citation', 'subject')],
                          'History': [('has_citation', 'subject')],
                          'LegalStatus': [('has_citation', 'subject')],
                          'Mandate': [('has_citation', 'subject')],
                          'Occupation': [('has_citation', 'subject')],
                          'PostalAddress': [],
                          'Structure': []})

    def test_scheme_tree(self):
        tree_def = container.tree_def(self.schema, 'ConceptScheme',
                                      skiprtypes=['wf_info_for']).items()
        tree_def = dict(sort_container(tree_def))
        self.assertEqual(tree_def,
                         {'Activity': [],
                          'Concept': [('generated', 'object'),
                                      ('label_of', 'object'),
                                      ('used', 'object')],
                          'ConceptScheme': [('generated', 'object'),
                                            ('in_scheme', 'object'),
                                            ('used', 'object')],
                          'Label': []})

    def test_concept_container(self):
        tree_def = container.tree_def(self.schema, 'Concept').items()
        tree_def = dict(sort_container(tree_def))
        self.assertEqual(tree_def,
                         {'Activity': [],
                          'Concept': [('generated', 'object'),
                                      ('label_of', 'object'),
                                      ('used', 'object')],
                          'Label': []})

    def test_seda_profile_tree(self):
        tree_def = container.tree_def(self.schema, 'SEDAProfile',
                                      skiprtypes=['wf_info_for']).items()
        tree_def = dict(sort_container(tree_def))
        self.assertEqual(tree_def,
                         {'ProfileArchiveObject': [('seda_access_restriction_code', 'subject'),
                                                   ('seda_appraisal_rule', 'subject'),
                                                   ('seda_content_description', 'subject'),
                                                   ('seda_name', 'subject'),
                                                   ('seda_parent', 'object')],
                          'ProfileDocument': [('seda_character_set_code', 'subject'),
                                              ('seda_description', 'subject'),
                                              ('seda_document_type_code', 'subject'),
                                              ('seda_file_type_code', 'subject'),
                                              ],
                          'SEDADocumentTypeCode': [],
                          'SEDACharacterSetCode': [],
                          'SEDAAccessRestrictionCode': [],
                          'SEDAAppraisalRule': [('seda_appraisal_rule_code', 'subject'),
                                                ('seda_appraisal_rule_duration', 'subject')],
                          'SEDAAppraisalRuleCode': [],
                          'SEDAAppraisalRuleDuration': [],
                          'SEDAContentDescription': [('seda_description', 'subject'),
                                                     ('seda_description_level', 'subject'),
                                                     ('seda_keyword_of', 'object'),
                                                     ('seda_latest_date', 'subject'),
                                                     ('seda_oldest_date', 'subject')],
                          'SEDADate': [],
                          'SEDADescription': [],
                          'SEDADescriptionLevel': [],
                          'SEDAFileTypeCode': [],
                          'SEDAKeyword': [],
                          'SEDAName': [],
                          'SEDAProfile': [('seda_parent', 'object')]})
        entity = self.vreg['etypes'].etype_class('SEDAProfile')(self)
        self.assertIsNotNone(entity.cw_adapt_to('ITree'))

    def test_lib_document_unit(self):
        with self.admin_access.repo_cnx() as cnx:
            testutils.seda_scheme(cnx, u'seda_document_type_code', u'preferred', u'CDO')
            name = cnx.create_entity('SEDAName')
            ar = cnx.create_entity('ProfileArchiveObject', seda_name=name)
            name = cnx.create_entity('SEDAName')
            ar2 = cnx.create_entity('ProfileArchiveObject', seda_name=name, seda_parent=ar)
            dctv = testutils.concept(cnx, 'CDO')
            dct = cnx.create_entity('SEDADocumentTypeCode', seda_document_type_code_value=dctv)
            doc = cnx.create_entity('ProfileDocument', seda_document_type_code=dct, seda_parent=ar)
            cnx.commit()
            self.assertEqual(ar.cw_adapt_to('IContained').container, None)
            itree = ar.cw_adapt_to('ITree')
            self.assertEqual(itree.parent(), None)
            self.assertEqual(set(e.eid for e in itree.children()), set([ar2.eid, doc.eid]))
            itree = ar2.cw_adapt_to('ITree')
            self.assertEqual(itree.parent().eid, ar.eid)
            self.assertEqual(itree.children(), [])
            itree = doc.cw_adapt_to('ITree')
            self.assertEqual(itree.parent().eid, ar.eid)
            self.assertEqual(itree.children(), [])

    def test_lib_document_unit_clone(self):
        """Functional test for SEDA component clone."""
        with self.admin_access.repo_cnx() as cnx:
            name = cnx.create_entity('SEDAName')
            ar = cnx.create_entity('ProfileArchiveObject', seda_name=name)
            name = cnx.create_entity('SEDAName', value=u'pouet')
            cloned = cnx.create_entity('ProfileArchiveObject', seda_name=name, seda_parent=ar)
            cnx.commit()
            cnx.create_entity(cloned.cw_etype, user_annotation=u'clone',
                              seda_parent=ar, seda_clone_of=cloned)
            cnx.commit()
            rset = cnx.execute(
                'Any X WHERE X user_annotation "clone", X seda_parent P')
            self.assertTrue(rset)

    def test_seda_profile_clone(self):
        """Functional test for SEDA profile cloning."""
        with self.admin_access.repo_cnx() as cnx:
            today = date.today()
            transferring_agent, archival_agent = testutils.setup_seda_agents(cnx)
            profile = cnx.create_entity('SEDAProfile', seda_transferring_agent=transferring_agent,
                                        ark_naa=testutils.naa(cnx))
            content_description = cnx.create_entity(
                'SEDAContentDescription',
                seda_description_level=cnx.create_entity('SEDADescriptionLevel'),
                seda_oldest_date=cnx.create_entity('SEDADate', value=today))
            restriction_code = cnx.create_entity('SEDAAccessRestrictionCode')
            seda_name = cnx.create_entity('SEDAName', value=u'Archive Name')
            cnx.create_entity('ProfileArchiveObject', seda_name=seda_name, seda_parent=profile,
                              seda_content_description=content_description,
                              seda_access_restriction_code=restriction_code)
            cnx.commit()
            profile.cw_adapt_to('IWorkflowable').fire_transition('publish')
            cnx.commit()
            clone = cnx.create_entity('SEDAProfile', title=u'Clone', seda_replace=profile,
                                      ark_naa=testutils.naa(cnx))
            cnx.commit()
            # ark and cwuri should not have been copied
            self.assertNotEqual(clone.ark, profile.ark)
            self.assertNotEqual(clone.cwuri, profile.cwuri)
            # Everything else should have been copied
            rql = ('Any AO WHERE AO seda_name SN, NOT SN eid %(sn_eid)s, SN value %(sn_value)s, '
                   'AO seda_content_description CD, CD seda_description_level DL, '
                   'CD seda_oldest_date OD, OD value %(od)s, AO seda_access_restriction_code RC, '
                   'AO seda_parent X, X title "Clone", NOT X eid %(original_eid)s')
            rset = cnx.execute(rql, {'sn_eid': seda_name.eid, 'sn_value': u'Archive Name',
                                     'od': today, 'original_eid': profile.eid})
            self.assertTrue(rset)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""Adapters to export content to RDF"""

from cubicweb.predicates import is_instance
from cubes.skos import rdfio
from cubes.skos.entities import AbstractRDFAdapter


def _register_agent_prov_mapping(reg):  # XXX move to the prov cube
    """Register RDF mapping for PROV-O entity types related to Agent.
    """
    reg.register_prefix('prov', 'http://www.w3.org/ns/prov#')
    # reg.register_etype_equivalence('Agent', 'prov:Agent')
    reg.register_etype_equivalence('Activity', 'prov:Activity')
    reg.register_attribute_equivalence('Activity', 'type', 'prov:type')
    reg.register_attribute_equivalence('Activity', 'description', 'prov:label')
    reg.register_attribute_equivalence('Activity', 'start', 'prov:startedAtTime')
    reg.register_attribute_equivalence('Activity', 'end', 'prov:endedAtTime')
    reg.register_relation_equivalence('Activity', 'associated_with', 'Agent',
                                      'prov:wasAssociatedWith')
    reg.register_relation_equivalence('Activity', 'generated', 'Agent', 'prov:generated')
    reg.register_relation_equivalence('Activity', 'used', 'Agent', 'prov:used')


def _base_rdf_mapping(reg, etype, foaf_kind):
    reg.register_prefix('foaf', 'http://xmlns.com/foaf/0.1/')
    reg.register_prefix('dc', 'http://purl.org/dc/elements/1.1/')
    reg.register_etype_equivalence(etype, 'foaf:' + foaf_kind)
    # Standard metadata.
    reg.register_attribute_equivalence(etype, 'name', 'foaf:name')
    # Additionnal metadata for Asalae export.
    reg.register_attribute_equivalence(etype, 'ark', 'dc:identifier')


def _complete_agent_rdf_mapping(reg, etype):
    reg.register_prefix('dct', 'http://purl.org/dc/terms/')
    reg.register_prefix('schema_org', 'http://schema.org/')
    reg.register_prefix('vcard', 'http://www.w3.org/2006/vcard/ns#')
    reg.register_attribute_equivalence(etype, 'creation_date', 'dct:created')
    reg.register_attribute_equivalence(etype, 'modification_date', 'dct:modified')


def _complete_authority_record_rdf_mapping(reg, etype):
    _complete_agent_rdf_mapping(reg, etype)
    reg.register_attribute_equivalence(etype, 'start_date', 'schema_org:startDate')
    reg.register_attribute_equivalence(etype, 'end_date', 'schema_org:endDate')
    reg.register_etype_equivalence('PostalAddress', 'vcard:Location')
    reg.register_attribute_equivalence('PostalAddress', 'street', 'vcard:street-address')
    reg.register_attribute_equivalence('PostalAddress', 'postalcode', 'vcard:postal-code')
    reg.register_attribute_equivalence('PostalAddress', 'city', 'vcard:locality')
    reg.register_attribute_equivalence('PostalAddress', 'country', 'vcard:country-name')
    reg.register_attribute_equivalence('PostalAddress', 'state', 'vcard:region')


class AgentRDFListAdapter(AbstractRDFAdapter):
    __regid__ = 'RDFList'
    __select__ = is_instance('Agent', 'OrganizationUnit')

    def register_rdf_mapping(self, reg):
        foaf_kind = {
            'Agent': 'Person',
            'OrganizationUnit': 'Organization',
        }[self.entity.cw_etype]
        _base_rdf_mapping(reg, self.entity.cw_etype, foaf_kind)


class AgentRDFPrimaryAdapter(AgentRDFListAdapter):
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('Agent')

    def register_rdf_mapping(self, reg):
        super(AgentRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _complete_agent_rdf_mapping(reg, self.entity.cw_etype)


class OrganizationUnitRDFPrimaryAdapter(AgentRDFListAdapter):
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('OrganizationUnit')

    def register_rdf_mapping(self, reg):
        super(OrganizationUnitRDFPrimaryAdapter, self).register_rdf_mapping(reg)
        _complete_agent_rdf_mapping(reg, self.entity.cw_etype)
        reg.register_relation_equivalence(
            'OrganizationUnit', 'contact_point', 'Agent',
            'schema_org:contactPoint')

    def fill(self, graph):
        super(OrganizationUnitRDFPrimaryAdapter, self).fill(graph)
        reg = self.registry
        # Export archival roles for the agent
        for archival_role in self.entity.archival_role:
            graph.add(graph.uri(self.entity.absolute_url()),
                      graph.uri(reg.normalize_uri('vcard:role')),
                      archival_role.name)
        # Export contact agent
        if self.entity.contact_point:
            contact_agent = self.entity.contact_point[0]
            contact_agent.cw_adapt_to('RDFPrimary').fill(graph)


class AuthorityRecordRDFPrimaryAdapter(AbstractRDFAdapter):
    __regid__ = 'RDFPrimary'
    __select__ = is_instance('AuthorityRecord')

    hierarchical_role = 'http://www.logilab.org/saem/hierarchical_role'
    association_role = 'http://www.logilab.org/saem/association_role'

    def register_rdf_mapping(self, reg):
        """RDF mapping for AuthorityRecord entity type."""
        reg.register_prefix('org', 'http://www.w3.org/ns/org#')
        foaf_kind = {'person': 'Person',
                     'authority': 'Organization',
                     'family': 'Group'}.get(self.entity.kind, 'Person')
        _base_rdf_mapping(reg, 'AuthorityRecord', foaf_kind)
        _complete_authority_record_rdf_mapping(reg, self.entity.cw_etype)

    def fill(self, graph):
        super(AuthorityRecordRDFPrimaryAdapter, self).fill(graph)
        reg = self.registry
        generator = rdfio.RDFGraphGenerator(graph)
        # Export addresses
        for place in self.entity.reverse_place_agent:
            for address in place.place_address:
                graph.add(graph.uri(self.entity.absolute_url()),
                          graph.uri(reg.normalize_uri('vcard:hasAddress')),
                          graph.uri(address.absolute_url()))
                if place.role:
                    graph.add(graph.uri(address.absolute_url()),
                              graph.uri(reg.normalize_uri('vcard:role')),
                              place.role)
                generator.add_entity(address, reg)
        # rtype name: (subj property, obj property)
        rtype_to_properties = {
            'reverse_chronological_successor': ('dct:replaces', 'dct:isReplacedBy'),
            'reverse_chronological_predecessor': ('dct:isReplacedBy', 'dct:replaces'),
            'reverse_hierarchical_parent': ('org:organization', 'org:member'),
            'reverse_hierarchical_child': ('org:member', 'org:organization'),
        }
        # Export chronological relations
        relations_generator = _iter_on_relations(
            self.entity,
            [('reverse_chronological_successor', 'chronological_predecessor'),
             ('reverse_chronological_predecessor', 'chronological_successor')])
        for rtype_name, relation, related_record in relations_generator:
            to_predicate, from_predicate = rtype_to_properties[rtype_name]
            graph.add(graph.uri(related_record.absolute_url()),
                      graph.uri(reg.normalize_uri(from_predicate)),
                      graph.uri(self.entity.absolute_url()))
            graph.add(graph.uri(self.entity.absolute_url()),
                      graph.uri(reg.normalize_uri(to_predicate)),
                      graph.uri(related_record.absolute_url()))
        # Export hierarchical relations
        relations_generator = _iter_on_relations(
            self.entity,
            [('reverse_hierarchical_parent', 'hierarchical_child'),
             ('reverse_hierarchical_child', 'hierarchical_parent')])
        self._add_membership_to_graph(graph, self.hierarchical_role, relations_generator,
                                      rtype_to_properties)
        # Export association relations
        relations_generator = _iter_on_relations(
            self.entity,
            [('reverse_association_from', 'association_to'),
             ('reverse_association_to', 'association_from')])
        # As this is an association relation, it makes no sense to have an agent being the
        # organization rather than the other one. That is why we "duplicate" Membership
        # properties: each agent will be both known as a member and an organization.
        relations = list(relations_generator)
        self._add_membership_to_graph(
            graph, self.association_role, relations,
            {'reverse_association_from': ('org:organization', 'org:member'),
             'reverse_association_to': ('org:member', 'org:organization')})
        self._add_membership_to_graph(
            graph, self.association_role, relations,
            {'reverse_association_from': ('org:member', 'org:organization'),
             'reverse_association_to': ('org:organization', 'org:member')})

    def _add_membership_to_graph(self, graph, role, relations, rtype_to_properties):
        reg = self.registry
        for rtype_name, relation, related_record in relations:
            relation_url = relation.absolute_url()
            graph.add(graph.uri(relation_url),
                      graph.uri(reg.normalize_uri(rtype_to_properties[rtype_name][0])),
                      graph.uri(self.entity.absolute_url()))
            graph.add(graph.uri(relation_url),
                      graph.uri(reg.normalize_uri(rtype_to_properties[rtype_name][1])),
                      graph.uri(related_record.absolute_url()))
            graph.add(graph.uri(relation_url),
                      graph.uri(reg.normalize_uri('org:role')),
                      graph.uri(role))
            if relation.start_date or relation.end_date:
                interval_resource = reg.normalize_uri(relation_url + '#timeInterval')
                graph.add(graph.uri(relation_url),
                          graph.uri(reg.normalize_uri('org:memberDuring')),
                          graph.uri(interval_resource))
            if relation.start_date:
                graph.add(graph.uri(interval_resource),
                          graph.uri(reg.normalize_uri('schema_org:startDate')),
                          relation.start_date)
            if relation.end_date:
                graph.add(graph.uri(interval_resource),
                          graph.uri(reg.normalize_uri('schema_org:endDate')),
                          relation.end_date)


def _iter_on_relations(entity, relation_descriptions):
    """yield (rtype_name, relation, related_entity)

    `relation_descriptions` is a list of (rtype_name, reverse_rtype_name)
    """
    for rtype_name, reverse_rtype_name in relation_descriptions:
        for relation in getattr(entity, rtype_name):
            for related_entity in getattr(relation, reverse_rtype_name):
                yield rtype_name, relation, related_entity



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem_ref unit tests for jqTree view"""
import json

from cubicweb.devtools.testlib import CubicWebTC

import testutils


def profile_document(cnx, **kwargs):
    """Return a complete ProfileDocument"""
    for rtype, label in [
        (u'seda_file_type_code', u'fmt/123'),
        (u'seda_character_set_code', u'6'),
        (u'seda_document_type_code', u'CDO'),
    ]:
        testutils.seda_scheme(cnx, rtype, u'preferred', label)
    cnx.commit()
    file_type_code_value = testutils.concept(cnx, 'fmt/123')
    file_type_code = cnx.create_entity(
        'SEDAFileTypeCode', user_cardinality=u'1',
        seda_file_type_code_value=file_type_code_value)
    character_set_code_value = testutils.concept(cnx, '6')
    character_set_code = cnx.create_entity(
        'SEDACharacterSetCode', user_cardinality=u'1',
        seda_character_set_code_value=character_set_code_value)
    document_type_code_value = testutils.concept(cnx, 'CDO')
    document_type_code = cnx.create_entity(
        'SEDADocumentTypeCode',
        seda_document_type_code_value=document_type_code_value)
    return cnx.create_entity(
        'ProfileDocument',
        seda_description=cnx.create_entity('SEDADescription'),
        seda_file_type_code=file_type_code,
        seda_character_set_code=character_set_code,
        seda_document_type_code=document_type_code,
        **kwargs
    )


def json_entity(entity, load_on_demand=False, selected=None):
    """Helper returning the minimal JSON output of an entity."""
    data = {
        u'id': entity.eid,
        u'type': entity.cw_etype,
        u'maybeChild': entity.cw_etype in (u'ProfileArchiveObject', u'ProfileDocument'),
        u'maybeParentOf': {
            u'SEDAProfile': [u'ProfileArchiveObject'],
            u'ProfileArchiveObject': [u'ProfileArchiveObject', u'ProfileDocument'],
            u'ProfileDocument': [],
        }[entity.cw_etype],
    }
    if load_on_demand:
        data[u'load_on_demand'] = True
    if selected:
        data[u'selected'] = selected
    return data


def ordered(jsobj):
    """Sort in place a JSON-like data structure."""
    if not jsobj:
        return
    if isinstance(jsobj, list):
        if isinstance(jsobj[0], basestring):
            jsobj.sort()
        elif isinstance(jsobj[0], dict):
            jsobj.sort(key=lambda x: x['id'])
            for item in jsobj:
                # make sure all items are dicts.
                assert isinstance(item, dict), item
                ordered(item)
    elif isinstance(jsobj, dict):
        jsobj.pop('label', None)
        for item in jsobj.itervalues():
            ordered(item)


class JQTreeViewTC(CubicWebTC):

    def _setup(self):
        with self.admin_access.repo_cnx() as cnx:
            profile = cnx.create_entity('SEDAProfile', ark_naa=testutils.naa(cnx))
            archobj1 = cnx.create_entity('ProfileArchiveObject', seda_parent=profile,
                                         seda_name=cnx.create_entity('SEDAName'))
            document1 = profile_document(cnx, seda_parent=archobj1)
            archobj2 = cnx.create_entity('ProfileArchiveObject', seda_parent=profile,
                                         seda_name=cnx.create_entity('SEDAName'))
            cnx.commit()
            return profile, archobj1, document1, archobj2

    def assertJsonEqual(self, actual, expected):
        ordered(actual)
        ordered(expected)
        self.assertEqual(actual, expected)

    def test_json(self):
        """jqtree.json view for entity."""
        profile, archobj1, document1, archobj2 = self._setup()
        with self.admin_access.web_request() as req:
            # profile level
            profile = req.entity_from_eid(profile.eid)
            archobj1 = req.entity_from_eid(archobj1.eid)
            archobj2 = req.entity_from_eid(archobj2.eid)
            # selected is `profile`.
            expected = json_entity(profile, selected=profile.eid)
            expected[u'children'] = [
                # archobj1 has load_on_demand, because it has a child itself.
                json_entity(archobj1, load_on_demand=True),
                json_entity(archobj2),
            ]
            self.assertJsonEqual(json.loads(profile.view('jqtree.json')),
                                 [expected])
            # archive object level
            archobj1 = req.entity_from_eid(archobj1.eid)
            # selected is `archobj1`.
            expected = json_entity(profile, selected=archobj1.eid)
            expected[u'children'] = [
                # archobj1 has load_on_demand, because it has a child itself.
                json_entity(archobj1, load_on_demand=True),
                json_entity(archobj2),
            ]
            self.assertJsonEqual(json.loads(archobj1.view('jqtree.json')),
                                 [expected])
            # document level
            doc1 = req.entity_from_eid(document1.eid)
            # selected is `doc1`.
            expected = json_entity(profile, selected=doc1.eid)
            expected[u'children'] = [
                # archobj1 has not load_on_demand, because its child is also
                # returned.
                json_entity(archobj1, load_on_demand=False),
                json_entity(archobj2),
            ]
            # append child of `archobj1`
            expected['children'][0]['children'] = [json_entity(doc1)]
            self.assertJsonEqual(json.loads(doc1.view('jqtree.json')),
                                 [expected])

    def test_json_node(self):
        """jqtree.json view with "node=eid" (returns children)."""
        profile, archobj1, document1, archobj2 = self._setup()
        with self.admin_access.web_request() as req:
            def call_view(eid):
                req.form['node'] = str(eid)
                return req.view('jqtree.json')
            # profile level
            archobj1 = req.entity_from_eid(archobj1.eid)
            archobj2 = req.entity_from_eid(archobj2.eid)
            expected = [
                # archobj1 has load_on_demand, because it has a child itself.
                json_entity(archobj1, load_on_demand=True),
                json_entity(archobj2),
            ]
            self.assertJsonEqual(json.loads(call_view(profile.eid)),
                                 expected)
            # archive object level
            archobj1 = req.entity_from_eid(archobj1.eid)
            doc1 = req.entity_from_eid(document1.eid)
            expected = [
                json_entity(doc1),
            ]
            self.assertJsonEqual(json.loads(call_view(archobj1.eid)),
                                 expected)
            # document level
            expected = []
            self.assertJsonEqual(json.loads(call_view(doc1.eid)),
                                 expected)


if __name__ == '__main__':
    import unittest
    unittest.main()



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref specific hooks and operations"""

from datetime import datetime

from logilab.common.registry import objectify_predicate

from yams import ValidationError
from yams.schema import role_name

from cubicweb.schema import META_RTYPES
from cubicweb.predicates import (is_instance, adaptable, on_fire_transition,
                                 relation_possible)
from cubicweb.server import hook
from cubicweb.hooks import metadata

from cubes.skos.hooks import ReplaceExternalUriByEntityHook
from cubes.saem_ref.entities import container

_ = unicode

# skip relation involved in the logging itself, or mirroring relation
IGNORE_RELATIONS = set(('used', 'generated', 'associated_with', 'mirror_of'))


@objectify_predicate
def contained_relation(cls, req, rtype, eidfrom, eidto, **kwargs):
    """Predicate that returns True for relation from/to a container or contained entity.
    """
    if rtype in IGNORE_RELATIONS:
        return 0
    for eid in (eidfrom, eidto):
        entity = req.entity_from_eid(eid)
        for interface in ('IContainer', 'IContained'):
            if entity.cw_adapt_to(interface):
                # skip meta relation of contained entities, only consider them on the container
                if interface == 'IContained' and rtype in META_RTYPES:
                    return 0
                return 1
    return 0


def qualify_relation(subj, rtype, obj):
    """Return a list of (relationship, role, target) qualifying the relation.

    Options are:

    * `relationship` = 'parent' and `target` the parent entity adapter
    * `relationship` = 'border' and `target` the container or contained adapter

    In both cases `role` is the role of the target in the relation.
    """
    # first attempt to get the more specific adapter for subject and object of the relation
    for interface in ('IContainer', 'IContained'):
        s_adapter = subj.cw_adapt_to(interface)
        if s_adapter is not None:
            break
    for interface in ('IContainer', 'IContained'):
        o_adapter = obj.cw_adapt_to(interface)
        if o_adapter is not None:
            break
    # then check if the relation is a parent relation
    if (s_adapter is not None and s_adapter.__regid__ == 'IContained'
            and (rtype, 'subject') in s_adapter.parent_relations):
        assert o_adapter, ('parent relation %s of contained entity #%d does not link to a '
                           'IContainer/IContained entity' % (rtype, subj.eid))
        return [('parent', 'object', o_adapter)]
    if (o_adapter is not None and o_adapter.__regid__ == 'IContained'
            and (rtype, 'object') in o_adapter.parent_relations):
        assert s_adapter, ('parent relation %s of contained entity #%d does not link to a '
                           'IContainer/IContained entity' % (rtype, obj.eid))
        return [('parent', 'subject', s_adapter)]
    # from here, we can assume rtype is not a parentship relation. Notice the relation may concerns
    # to distinct container
    assert s_adapter or o_adapter
    result = []
    if s_adapter is not None:
        result.append(('border', 'subject', s_adapter))
    if o_adapter is not None:
        result.append(('border', 'object', o_adapter))
    return result


# generic hooks to record operations if something changed in a compound tree ###

class AddOrRemoveChildrenHook(hook.Hook):
    """Some relation involved in a compound graph is added or removed."""
    __regid__ = 'compound.graph.updated'
    __select__ = hook.Hook.__select__ & contained_relation()
    events = ('before_add_relation', 'before_delete_relation')
    category = 'metadata'

    def __call__(self):
        for relationship, role, target in qualify_relation(self._cw.entity_from_eid(self.eidfrom),
                                                           self.rtype,
                                                           self._cw.entity_from_eid(self.eidto)):
            entity = target.entity
            UpdateModificationDateOp.get_instance(self._cw).add_entity(entity)
            if entity.cw_etype in container.PROFILE_CONTAINER_DEF:
                CheckProfileSEDACompatiblityOp.get_instance(self._cw).add_entity(entity)


class UpdateEntityHook(hook.Hook):
    """Some entity involved in a compound graph is updated."""
    __regid__ = 'compound.graph.updated'
    __select__ = hook.Hook.__select__ & adaptable('IContainer', 'IContained')
    events = ('before_update_entity',)
    category = 'metadata'

    def __call__(self):
        UpdateModificationDateOp.get_instance(self._cw).add_entity(self.entity)
        if self.entity.cw_etype in container.PROFILE_CONTAINER_DEF:
            CheckProfileSEDACompatiblityOp.get_instance(self._cw).add_entity(self.entity)


# Auto-update of modification date in a compound tree ##########################

class UpdateModificationDateOp(hook.DataOperationMixIn, hook.Operation):
    """Data operation updating the modification date of its data entities."""

    def precommit_event(self):
        cnx = self.cnx
        now = datetime.utcnow()
        with cnx.deny_all_hooks_but():
            for eid in self.get_data():
                if cnx.deleted_in_transaction(eid) or cnx.added_in_transaction(eid):
                    continue
                entity = cnx.entity_from_eid(eid)
                entity.cw_set(modification_date=now)

    def add_entity(self, entity):
        """Add entity, its parent entities (up to the container root) for update of their
        modification date at commit time.
        """
        self.add_data(entity.eid)
        while True:
            safety_belt = set((entity.eid,))
            contained = entity.cw_adapt_to('IContained')
            if contained is None:
                assert entity.cw_adapt_to('IContainer')
                break
            else:
                entity = contained.parent
                if entity is None or entity.eid in safety_belt:
                    break
                self.add_data(entity.eid)
                safety_belt.add(entity.eid)


# Transformation of ExternalUri to AuthorityRecord

class ReplaceExternalUriByAuthorityRecordHook(ReplaceExternalUriByEntityHook):
    """Replace ExternalUri by an AuthorityRecord"""
    __select__ = ReplaceExternalUriByEntityHook.__select__ & is_instance('AuthorityRecord')


# ARK generation ###############################################################

class AssignARKHook(hook.Hook):
    """When an entity supporting ARK is created, assign it a local ark to."""
    __regid__ = 'saem_ref.ark.assign'
    __select__ = hook.Hook.__select__ & relation_possible('ark')
    events = ('before_add_entity',)
    order = metadata.InitMetaAttrsHook.order - 1

    def __call__(self):
        if not self.entity.cw_edited.get('ark'):
            naa_what = self.entity.cw_adapt_to('IArkNAALocator').naa_what()
        else:
            naa_what = None
        set_ark_and_cwuri(self._cw, self.entity.eid, self.entity.cw_etype,
                          self.entity.cw_edited, naa_what=naa_what)


def set_ark_and_cwuri(cw, eid, etype, entity_attrs, naa_what=None):
    ark = entity_attrs.get('ark')
    if not ark:
        cwuri = entity_attrs.get('cwuri')
        ark = None if cwuri is None else extract_ark(cwuri)
        if ark is None:
            if naa_what is None:
                msg = _('an ARK identifier has to be generated but no Name Assigning Authority is '
                        'specified')
                raise ValidationError(None, {'': msg})
            generator = cw.vreg['adapters'].select('IARKGenerator', cw, eid=eid, etype=etype,
                                                   naa_what=naa_what)
            ark = generator.generate_ark()
        entity_attrs['ark'] = ark
    if 'cwuri' not in entity_attrs:
        # store ark as cwuri, not an URL so it's easier to move the database while still easy to get
        # an URL from there (see the cwuri_url function) XXX (syt) any other reason to do so? there
        # is also probably some constraint on (re)import or something similar but I don't recall
        # right now
        entity_attrs['cwuri'] = u'ark:/' + ark


def extract_ark(url):
    """Extract ARK identifier from an URL, return it or None if not found or malformed.
    """
    try:
        _, ark = url.split('ark:/', 1)
    except ValueError:
        return None
    parts = ark.split('/')
    if len(parts) < 2:
        return None
    ark = '/'.join(parts[:2])
    for delim in '#?':
        ark = ark.split(delim, 1)[0]
    return ark


# Life-cycle logging ###########################################################

class Record(object):
    """Temporary representation of some activities to be recorded, for accumulation per entity
    prior to merge in an operation.
    """
    def __init__(self, events):
        if not isinstance(events, set):
            events = set(events)
        self.events = events

    def __str__(self):
        return ','.join(sorted('%s %s' % ev for ev in self.events))

    def merge(self, other):
        """Merge the activity with another."""
        self.events |= other.events

    def as_dict(self, _):
        """Return a dictionary suitable for creation of the Activity entity."""
        now = datetime.utcnow()
        record = {'start': now, 'end': now}
        # merge events on the same target
        targets_per_event_type = {}
        events_by_target = {}
        for ev_type, ev_target in self.events:
            events_by_target.setdefault(ev_target, set([])).add(ev_type)
        for target, target_events in events_by_target.iteritems():
            # if there are several event types for the same target (e.g. added and removed a
            # relation), group them into the "modified" event type
            if len(target_events) > 1:
                ev_type = 'modified'
            else:
                ev_type = iter(target_events).next()
            targets_per_event_type.setdefault(_(ev_type), []).append(target)
        # now generate proper messages for each event type
        msgs = []
        # iterates on event types to get expected order
        for ev_type in (_('created'), _('modified'), _('added'), _('removed')):
            if ev_type not in targets_per_event_type:
                continue
            modified = sorted(_(x).lower() for x in targets_per_event_type[ev_type])
            msgs.append(u'%s %s' % (ev_type, u', '.join(modified)))
        if len(msgs) == 1:
            record['description'] = msgs[0]
        else:
            record['description'] = u'\n'.join(u'* ' + msg for msg in msgs)
        # complete record with other metadata then we're done
        record['description_format'] = u'text/rest'
        record['type'] = u'create' if _('created') in targets_per_event_type else u'modify'
        return record


class AddActivityOperation(hook.DataOperationMixIn, hook.LateOperation):
    """The operation responsible to merge activites to their container then to record them."""
    def precommit_event(self):
        # translate using site default language
        lang = self.cnx.vreg.property_value('ui.language')
        _ = self.cnx.vreg.config.translations[lang][0]
        # first, merge all activities (there may be several for the same container entity)
        activity_per_container = {}
        for entity, activity in self.get_data():
            # find the entity on which we should log the activity
            icontainer = entity.cw_adapt_to('IContainer')
            if icontainer is None:
                icontainer = entity.cw_adapt_to('IContained')
                if icontainer is None:
                    # "free" entity, no logging
                    self.warning("%s is not in a container, don't record activity", entity)
                    continue
            container = icontainer.container
            # if the container is not yet linked or has been deleted in this transaction, we've
            # nothing to log on
            if container is None or self.cnx.deleted_in_transaction(container.eid):
                continue
            # if the container has been created in this transaction, discard all activities but the
            # creation
            if self.cnx.added_in_transaction(container.eid):
                if ('created', entity.cw_etype) in activity.events:
                    # only keep the creation event
                    assert container not in activity_per_container
                    activity_per_container[container] = activity
            else:
                try:
                    activity_per_container[container].merge(activity)
                except KeyError:
                    activity_per_container[container] = activity
        # who?
        if self.cnx.user.eid == -1:  # internal manager
            agent = None
        else:
            user = self.cnx.entity_from_eid(self.cnx.user.eid)
            agent = user.reverse_agent_user and user.reverse_agent_user[0] or None
        # then record the activities
        for container, activity in activity_per_container.items():
            kwargs = activity.as_dict(_)
            if agent is not None:
                kwargs.setdefault('associated_with', agent)
            adapted = container.cw_adapt_to('IRecordable')
            if adapted is not None:
                adapted.add_activity(**kwargs)


class LogContainerCreationHook(hook.Hook):
    """Add an Activity upon creation of a container entity (`IContainer`). There is no need for a
    hook on `IContained` creation, we'll catch creation of the relation linking them to their
    container.
    """
    __select__ = hook.Hook.__select__ & adaptable('IContainer')
    __regid__ = 'saem_ref.log.add'
    events = ('after_add_entity',)
    category = 'prov.logging'

    def __call__(self):
        activity = Record([('created', self.entity.cw_etype)])
        AddActivityOperation.get_instance(self._cw).add_data((self.entity, activity))


class LogModificationHook(hook.Hook):
    """Add an Activity upon modification of a container (`IContainer`) or contained (`IContained`)
    entity. The message should be adapted accordingly.
    """
    __select__ = hook.Hook.__select__ & adaptable('IContainer', 'IContained')
    __regid__ = 'saem_ref.log.update'
    events = ('after_update_entity',)
    category = 'prov.logging'

    def __call__(self):
        entity = self.entity
        if entity.cw_adapt_to('IContainer'):
            attrs = set(attr for attr in entity.cw_edited if attr not in META_RTYPES)
            if not attrs:
                return
            events = [('modified', attr) for attr in attrs]
        else:  # IContained
            parent_relation = entity.cw_adapt_to('IContained').parent_relation()
            if parent_relation is None:
                # not yet bound to a parent, skip this activity (later addition of the relation will
                # trigger the record if necessary)
                return
            rtype, role = parent_relation
            # role is the role of the entity, we want to qualify the relation by the role of the
            # parent
            if role == 'subject':
                rtype += '_object'
            events = [('modified', rtype)]
        activity = Record(events)
        AddActivityOperation.get_instance(self._cw).add_data((entity, activity))


class LogRelationHook(hook.Hook):
    """Add an Activity upon modification of a container (`IContainer`) or contained (`IContained`)
    entity. The message should be adapted accordingly.
    """
    __select__ = hook.Hook.__select__ & contained_relation()
    __regid__ = 'saem_ref.log.relation'
    events = ('after_add_relation', 'before_delete_relation')
    category = 'prov.logging'

    def __call__(self):
        for relationship, role, target in qualify_relation(self._cw.entity_from_eid(self.eidfrom),
                                                           self.rtype,
                                                           self._cw.entity_from_eid(self.eidto)):
            rtype = self.rtype
            if role == 'object':
                rtype += '_object'
            if self.event == 'after_add_relation':
                events = [('added', rtype)]
                self.add_activity(target.entity, events)
            else:  # before_delete_relation
                events = [('removed', rtype)]
                # in the case of a relation deletion, we've to find the container right away without
                # waiting for the operation since the parent relation may have been removed at this
                # point
                container = target.container
                if container is not None:
                    self.add_activity(container, events)

    def add_activity(self, entity, events):
        activity = Record(events)
        AddActivityOperation.get_instance(self._cw).add_data((entity, activity))


# SEDA ###############################################################

class CheckProfileSEDACompatiblityOp(hook.DataOperationMixIn, hook.Operation):
    """Data operation that will check compatibility of a SEDA profile upon modification."""
    seda_adapters = ['SEDA-0.2.xsd', 'SEDA-1.0.xsd']

    def precommit_event(self):
        cnx = self.cnx
        profiles = set()
        for eid in self.get_data():
            if cnx.deleted_in_transaction(eid):
                continue
            entity = cnx.entity_from_eid(eid)
            if entity.cw_etype == 'SEDAProfile':
                profiles.add(entity)
            else:
                container = entity.cw_adapt_to('IContained').container
                if container is not None:
                    profiles.add(container)
        with cnx.deny_all_hooks_but():
            for profile in profiles:
                supported = []
                for adapter_id in self.seda_adapters:
                    if profile.cw_adapt_to(adapter_id).is_compatible():
                        supported.append(adapter_id)
                profile.cw_set(support_seda_exports=u', '.join(sorted(supported)))

    def add_entity(self, entity):
        """Add entity, its parent entities (up to the container root) for update of their
        modification date at commit time.
        """
        self.add_data(entity.eid)
        while True:
            safety_belt = set((entity.eid,))
            contained = entity.cw_adapt_to('IContained')
            if contained is None:
                assert entity.cw_adapt_to('IContainer')
                break
            else:
                entity = contained.container
                if entity is None or entity.eid in safety_belt:
                    break
                self.add_data(entity.eid)
                safety_belt.add(entity.eid)


class CheckNewProfile(hook.Hook):
    """Instantiate operation checking for profile seda compat on its creation"""
    __regid__ = 'saem_ref.create-profile'
    __select__ = hook.Hook.__select__ & is_instance('SEDAProfile')
    events = ('after_add_entity', )

    def __call__(self):
        CheckProfileSEDACompatiblityOp.get_instance(self._cw).add_entity(self.entity)


class AddMandatoryRelationsToProfileArchiveHook(hook.Hook):
    """Create mandatory relations upon creation of first level ProfileArchiveObject (easing
    entity creation from the UI).
    """
    __select__ = hook.Hook.__select__ & hook.match_rtype('seda_parent', toetypes=('SEDAProfile',))
    __regid__ = 'seda.add_mandatory_relations_to_profilearchive'
    events = ('after_add_relation',)

    def __call__(self):
        document_unit = self._cw.entity_from_eid(self.eidfrom)
        if not document_unit.access_restriction_code:
            self._cw.create_entity('SEDAAccessRestrictionCode',
                                   user_cardinality=u'1',
                                   reverse_seda_access_restriction_code=document_unit)
        if not document_unit.content_description:
            desc_level = self._cw.create_entity('SEDADescriptionLevel')
            self._cw.create_entity('SEDAContentDescription',
                                   user_cardinality=u'1',
                                   seda_description_level=desc_level,
                                   reverse_seda_content_description=document_unit)


class DepreciateProfileOnReplacePublishedHook(hook.Hook):
    """Depreciate a SEDA profile once the new version is published."""
    __select__ = hook.Hook.__select__ & on_fire_transition('SEDAProfile', 'publish')
    __regid__ = 'seda.deprecate_profile_on_replace_published'
    events = ('after_add_entity',)

    def __call__(self):
        replaced = self.entity.for_entity.seda_replace
        if replaced:
            wf = replaced[0].cw_adapt_to('IWorkflowable')
            wf.fire_transition_if_possible("deprecate")


# Relations deposit agent - archival agent  ####################################

class DontDeleteUnitIfArchival(hook.Hook):
    """Before deleting an agent, if it is an archival agent, make sure it isn't
    associated to an organization through the "archival_unit" relation.
    """
    __select__ = hook.Hook.__select__ & hook.match_rtype('archival_role')
    __regid__ = 'saem_ref.archival_role.delete'
    events = ('before_delete_relation', )

    def __call__(self):
        ou = self._cw.entity_from_eid(self.eidfrom)
        role = self._cw.entity_from_eid(self.eidto)
        if role.name == u'archival' and ou.reverse_archival_unit:
            errors = {role_name('archival_role', 'subject'):
                      _('this organization unit is the archival unit of some organization, '
                        'therefore the role "archival" cannot be deleted')}
            raise ValidationError(ou.eid, errors)


# ensure that equivalent_concept Concept has vocabulary_source defined ####################

class EnsureVocabularySource(hook.Hook):
    """When a equivalent_concept relation is set and targets a Concept, ensure that the
    vocabulary_source relation is set to the concept's scheme. This should not be necessary when
    using the UI where the workflow enforce setting the scheme first, but it's necessary during
    e.g. EAC import.
    """
    __select__ = hook.Hook.__select__ & hook.match_rtype('equivalent_concept',
                                                         toetypes=('Concept',))
    __regid__ = 'saem_ref.add_equivalent_concept'
    events = ('after_add_relation', )

    def __call__(self):
        EnsureVocabularySourceOp.get_instance(self._cw).add_data((self.eidfrom, self.eidto))


class EnsureVocabularySourceOp(hook.DataOperationMixIn, hook.Operation):
    """Ensure X equivalent_concept target Concept as proper X vocabulary_source Scheme"""

    def precommit_event(self):
        cnx = self.cnx
        for eid, concept_eid in self.get_data():
            cnx.execute('SET X vocabulary_source SC WHERE NOT X vocabulary_source SC, '
                        'C in_scheme SC, C eid %(c)s, X eid %(x)s', {'x': eid, 'c': concept_eid})


def registration_callback(vreg):
    vreg.register_all(globals().values(), __name__)
    from cubicweb.server import ON_COMMIT_ADD_RELATIONS
    from cubes.saem_ref import PERMISSIONS_GRAPHS, OUTER_RELATIVE_PERM_RELATIONS, mandatory_rdefs
    # Add relations involved in a composite graph with security setup to "on
    # commit" check step.
    schema = vreg.schema
    for etype, graph_def in PERMISSIONS_GRAPHS.iteritems():
        graph = graph_def(schema)
        for rdef, _ in mandatory_rdefs(schema, graph.parent_structure(etype)):
            ON_COMMIT_ADD_RELATIONS.add(rdef.rtype)
    # As well as those whose security rules are defined in the schema directly.
    ON_COMMIT_ADD_RELATIONS.update(OUTER_RELATIVE_PERM_RELATIONS)



# copyright 2015-2016 LOGILAB S.A. (Paris, FRANCE), all rights reserved.
# contact http://www.logilab.fr -- mailto:contact@logilab.fr
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU Lesser General Public License as published by the Free
# Software Foundation, either version 2.1 of the License, or (at your option)
# any later version.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License for more
# details.
#
# You should have received a copy of the GNU Lesser General Public License along
# with this program. If not, see <http://www.gnu.org/licenses/>.
"""cubicweb-saem-ref views related to SEDA"""

from copy import copy, deepcopy
from functools import partial
from collections import defaultdict

from logilab.mtconverter import xml_escape

from cubicweb import tags, uilib, utils
from cubicweb.view import EntityView
from cubicweb.predicates import (is_instance, score_entity, one_line_rset, multi_lines_rset,
                                 has_permission, match_form_params, has_related_entities,
                                 is_in_state, relation_possible)
from cubicweb.web import (Redirect, httpcache, action, component, controller,
                          formwidgets as fw, formfields as ff)
from cubicweb.web.views import (ibreadcrumbs, idownloadable, uicfg, baseviews, basecomponents,
                                tableview, primary,
                                formrenderers, actions, editforms)

from cubes.relationwidget import views as rwdg
from cubes.compound.entities import copy_entity
from cubes.compound.views import CloneAction

from cubes.saem_ref.views import jqtree
from cubes.saem_ref.views import add_etype_link, configure_relation_widget, index_dropdown_button
from cubes.saem_ref.views.widgets import ConceptAutoCompleteWidget


_ = unicode


def selectable_uicfg(rtag, modname, select):
    """Return a copy of uicfg `rtag` with `select` set.

    Usually `modname` should be set to ``__name__`` from the caller.
    """
    copy = deepcopy(rtag)
    copy.__module__ = modname
    copy.__select__ = select
    return copy


# custom widget ####################################################################################

class SEDAMetaFieldMixIn(object):
    """This mix-in is designed to create compound field allowing edition of a `user_annotation` and
    optionaly a `user_cardinality` fields, which appear hidden by default.
    """

    @staticmethod
    def cardinality_field(form):
        """Return the field for `user_cardinality` if the attribute is applyable to the edited
        entity.
        """
        eschema = form.edited_entity.e_schema
        if 'user_cardinality' in eschema.subjrels:
            rschema = eschema.subjrels['user_cardinality']
            return ff.guess_field(eschema, rschema, 'subject', form._cw)
        return None

    @staticmethod
    def annotation_field(form):
        """Return the field for `user_annotation` entity. """
        eschema = form.edited_entity.e_schema
        rschema = eschema.subjrels['user_annotation']
        return ff.guess_field(eschema, rschema, 'subject', form._cw)

    def actual_fields(self, form):
        """Overriden from :class:`Field`"""
        yield self
        card_field = self.cardinality_field(form)
        if card_field:
            yield card_field
        yield self.annotation_field(form)

    def render(self, form, renderer):
        """Overriden from :class:`Field` to render fields in a div which is hidden by default"""
        form._cw.add_js('cubes.saem_ref.js')
        wdgs = [self.get_widget(form).render(form, self, renderer)]
        self._render_hidden_section(wdgs, form, renderer)
        return u'\n'.join(wdgs)

    def _render_hidden_section(self, wdgs, form, renderer):
        divid = '%s-advanced' % self.input_name(form)
        self._render_hidden_section_title(wdgs, form, divid)
        wdgs.append(u'<div id="%s" class="hidden">' % divid)
        card_field = self.cardinality_field(form)
        if card_field:
            wdgs.append(self._render_subfield(form, card_field, renderer))
        wdgs.append(self._render_subfield(form, self.annotation_field(form), renderer))
        wdgs.append(u'</div>')

    @staticmethod
    def _render_hidden_section_title(wdgs, form, divid):
        wdgs.append(tags.a(u'', onclick=unicode(uilib.js.saem.toggleFormMetaVisibility(divid)),
                           href='javascript:$.noop()', title=form._cw._('show meta fields'),
                           klass='icon-list-add'))

    @staticmethod
    def _render_subfield(form, field, renderer):
        """Render a sub-field: label + widget + help + EOL"""
        data = utils.UStringIO()
        w = data.write
        w(u'<div class="row">')
        w(renderer.render_label(form, field))
        w(u'<div class="col-md-9">')
        w(field.render(form, renderer))
        w(u'</div>')
        w(u'</div>')
        w(u'<div class="row">')
        w(u'<div class="col-md-offset-3 col-md-9">')
        w(renderer.render_help(form, field))
        w(u'</div>')
        w(u'</div>')
        return data.getvalue()

    def get_widget(self, form):
        """return the widget instance associated to this field"""
        widget = super(SEDAMetaFieldMixIn, self).get_widget(form)
        widget.attrs['class'] = 'metaField'
        return self.widget


class SEDAMetaField(SEDAMetaFieldMixIn, ff.Field):
    """Field to handle case where the cardinality and annotation fields have not field to which to
    be attached to.
    """

    def __init__(self, *args, **kwargs):
        super(SEDAMetaField, self).__init__(*args, **kwargs)
        # XXX need a space so an empty label is generated and so we get proper fields alignment
        self.label = ' '

    def actual_fields(self, form):
        """Overriden from :class:`SEDAMetaFieldMixIn`"""
        card_field = self.cardinality_field(form)
        if card_field:
            yield card_field
        yield self.annotation_field(form)

    def render(self, form, renderer):
        """Overriden from :class:`Field` to render fields in a div which is hidden by default"""
        form._cw.add_js('cubes.saem_ref.js')
        wdgs = []
        self._render_hidden_section(wdgs, form, renderer)
        return u'\n'.join(wdgs)

    # def _render_hidden_section_title(self, wdgs, form, divid):
    #     wdgs.append(tags.a(form._cw._('show meta fields'),
    #                        onclick=unicode(uilib.js.saem.toggleFormMetaVisibility(divid)),
    #                        href='javascript:$.noop()'))


# primary view configuration #######################################################################

afs = uicfg.autoform_section
pvs = uicfg.primaryview_section
pvds = uicfg.primaryview_display_ctrl

pvs.tag_subject_of(('*', 'seda_appraisal_rule', '*'), 'attributes')
pvs.tag_subject_of(('*', 'seda_access_restriction_code', '*'), 'attributes')
pvs.tag_subject_of(('*', 'seda_content_description', '*'), 'attributes')
pvs.tag_subject_of(('*', 'seda_oldest_date', '*'), 'attributes')
pvs.tag_subject_of(('*', 'seda_latest_date', '*'), 'attributes')
pvs.tag_subject_of(('*', 'seda_description', '*'), 'attributes')
pvs.tag_object_of(('*', 'seda_keyword_of', '*'), 'attributes')
pvs.tag_object_of(('*', 'seda_parent', '*'), 'hidden')
pvs.tag_subject_of(('*', 'seda_document_type_code', '*'), 'attributes')
pvs.tag_subject_of(('*', 'seda_file_type_code', '*'), 'attributes')
pvs.tag_subject_of(('*', 'seda_character_set_code', '*'), 'attributes')
pvs.tag_attribute(('SEDAProfile', 'ark'), 'attributes')

# related SKOS object
pvs.tag_object_of(('*', 'seda_keyword_scheme', '*'), 'hidden')
pvs.tag_subject_of(('ConceptScheme', 'scheme_relation', '*'), 'hidden')
afs.tag_subject_of(('ConceptScheme', 'scheme_relation', '*'), 'main', 'hidden')

# hide default relation components for 'seda_replace' rtype
pvs.tag_subject_of(('*', 'seda_replace', '*'), 'hidden')
pvs.tag_object_of(('*', 'seda_replace', '*'), 'hidden')

# reledit (primary view) configuration #############################################################

rec = uicfg.reledit_ctrl
rec.tag_subject_of(('*', 'seda_name', '*'),
                   {'rvid': 'saem.seda.value'})
rec.tag_subject_of(('*', 'seda_description', '*'),
                   {'rvid': 'saem.seda.value'})
rec.tag_subject_of(('*', 'seda_access_restriction_code', '*'),
                   {'rvid': 'saem.seda.access-restriction-value'})
rec.tag_subject_of(('*', 'seda_appraisal_rule', '*'),
                   {'rvid': 'saem.seda.appraisal-rule-value'})
rec.tag_subject_of(('*', 'seda_content_description', '*'),
                   {'rvid': 'saem.seda.content-description-value'})
rec.tag_subject_of(('*', 'seda_document_type_code', '*'),
                   {'rvid': 'saem.seda.document-type-code-value'})
rec.tag_subject_of(('*', 'seda_file_type_code', '*'),
                   {'rvid': 'saem.seda.file-type-code-value'})
rec.tag_subject_of(('*', 'seda_character_set_code', '*'),
                   {'rvid': 'saem.seda.character-set-code-value'})


class ValueView(EntityView):
    __regid__ = 'saem.seda.value'

    def entity_call(self, entity):
        if entity.value:
            self.wdata(entity.value)
        else:
            self.wdata(self._cw._('<no value>'))


class AccessRestrictionValueView(EntityView):
    __regid__ = 'saem.seda.access-restriction-value'

    def entity_call(self, entity):
        if entity.seda_access_restriction_code_value:
            entity.seda_access_restriction_code_value[0].view('oneline', w=self.w)
        else:
            self.wdata(self._cw._('<no value>'))


class SEDAFileTypeCodeValueView(EntityView):
    __regid__ = 'saem.seda.file-type-code-value'

    def entity_call(self, entity):
        if entity.seda_file_type_code_value:
            entity.seda_file_type_code_value[0].view('saem.concept.alt', w=self.w)
        else:
            self.wdata(self._cw._('<no value>'))


class SEDADocumentTypeCodeValueView(EntityView):
    __regid__ = 'saem.seda.document-type-code-value'

    def entity_call(self, entity):
        if entity.seda_document_type_code_value:
            entity.seda_document_type_code_value[0].view('oneline', w=self.w)
        else:
            self.wdata(self._cw._('<no value>'))


class SEDACharacterSetCodeValueView(EntityView):
    __regid__ = 'saem.seda.character-set-code-value'

    def entity_call(self, entity):
        if entity.seda_character_set_code_value:
            entity.seda_character_set_code_value[0].view('saem.concept.alt', w=self.w)
        else:
            self.wdata(self._cw._('<no value>'))


class AppraisalRuleValueView(EntityView):
    __regid__ = 'saem.seda.appraisal-rule-value'

    def entity_call(self, entity):
        _ = self._cw._
        if (entity.seda_appraisal_rule_duration and
                entity.appraisal_rule_duration.seda_appraisal_rule_duration_value):
            rule = entity.appraisal_rule_duration
            duration = rule.seda_appraisal_rule_duration_value[0].view('saem.concept.alt')
        else:
            duration = _('<no value>')
        self.w('<div><span>%s</span> %s</div>' % (_('seda_appraisal_rule_duration'), duration))
        if (entity.seda_appraisal_rule_code and
                entity.appraisal_rule_code.seda_appraisal_rule_code_value):
            rule = entity.appraisal_rule_code
            code = rule.seda_appraisal_rule_code_value[0].view('oneline')
        else:
            code = xml_escape(_('<no value>'))
        self.w('<div><span>%s</span> %s</div>' % (_('seda_appraisal_rule_code'), code))


class ContentDescriptionValueView(EntityView):
    __regid__ = 'saem.seda.content-description-value'

    def entity_call(self, entity):
        _ = self._cw._
        html = u''
        if entity.seda_description_level[0].seda_description_level_value:
            level = entity.seda_description_level[0].seda_description_level_value[0].view('oneline')
            html += '<div><span>%s</span> %s</div>' % (_('seda_description_level'), level)
        for field in ('description', 'oldest_date', 'latest_date'):
            value_entity = getattr(entity, 'seda_' + field)
            if value_entity:
                value_entity = value_entity[0]
                value = (value_entity.printable_value('value')
                         if value_entity.value else _('<no value>'))
                html += '<div><span>%s</span> %s</div>' % (self._cw.__('seda_' + field),
                                                           xml_escape(value))
        if entity.keywords:
            keywords = self._cw.view('csv', entity.related('seda_keyword_of', 'object'))
            html += '<div><span>%s</span> %s</div>' % (_('seda_keyword_of_object'), keywords)
        self.w(html or xml_escape(_(u'<no value>')))


# forms configuration ##############################################################################


def rtype_concept_label_choices(form, field, label_kind=u'preferred_label'):
    """To be used as a ``choices`` function for an attribute whose objects are SKOS concepts.

    With this function, a combo box will be displayed with a list of labels for all relevant
    concepts. Kind of displayed labels is given by the `label_kind` parameter; by default, preferred
    labels are displayed.
    """
    assert field.name.endswith('_value')
    rtype_name = unicode(field.name[:-6])
    rql = ('Any LABEL,C WHERE S scheme_relation RT, C in_scheme S, C {0} L,'
           ' L label LABEL, RT name %(rtype_name)s').format(label_kind)
    # XXX: if a concept has multiple labels of given kind, they will all appear in the list.
    # This should not happen since these concepts are created programmatically, but a user may edit
    rset = form._cw.execute(rql, {'rtype_name': rtype_name})
    return [(label, unicode(concept_eid)) for label, concept_eid in rset.rows]

rtype_concept_alt_label_choices = partial(rtype_concept_label_choices,
                                          label_kind=u'alternative_label')
rtype_concept_pref_label_choices = partial(rtype_concept_label_choices,
                                           label_kind=u'preferred_label')


affk = uicfg.autoform_field_kwargs
aff = uicfg.autoform_field

# name (for archive and archive object)
afs.tag_subject_of(('*', 'seda_name', '*'), 'main', 'inlined')
affk.set_field_kwargs('SEDAName', 'value', widget=fw.TextInput({'size': 80}))

# description (for document and content description)
afs.tag_subject_of(('*', 'seda_description', '*'), 'main', 'inlined')

# content description (for archive and archive object)
afs.tag_subject_of(('*', 'seda_content_description', '*'), 'inlined', 'hidden')
afs.tag_subject_of(('*', 'seda_content_description', '*'), 'main', 'inlined')
afs.tag_object_of(('*', 'seda_keyword_of', '*'), 'main', 'inlined')
afs.tag_subject_of(('*', 'seda_description_level', '*'), 'main', 'inlined')
afs.tag_subject_of(('*', 'seda_oldest_date', '*'), 'main', 'inlined')
afs.tag_subject_of(('*', 'seda_latest_date', '*'), 'main', 'inlined')
afs.tag_subject_of(('SEDAContentDescription', 'seda_description', '*'), 'main', 'inlined')
afs.tag_object_of(('*', 'seda_keyword_of', 'SEDAContentDescription'),
                  'main', 'inlined')
afs.tag_subject_of(('SEDAContentDescription', 'seda_description_level', '*'),
                   'main', 'inlined')
afs.tag_subject_of(('SEDADescriptionLevel', 'seda_description_level_value', '*'),
                   'main', 'attributes')
afs.tag_subject_of(('SEDAContentDescription', 'seda_oldest_date', '*'),
                   'main', 'inlined')
afs.tag_subject_of(('SEDAContentDescription', 'seda_latest_date', '*'),
                   'main', 'inlined')
afs.tag_subject_of(('SEDAKeyword', 'seda_keyword_scheme', '*'),
                   'main', 'attributes')
afs.tag_subject_of(('SEDAKeyword', 'seda_keyword_value', '*'),
                   'main', 'attributes')

affk.set_field_kwargs('SEDAKeyword', 'seda_keyword_value',
                      widget=ConceptAutoCompleteWidget(slave_name='keyword_value',
                                                       master_name='keyword_scheme'))

# appraisal rule (for archive and archive object)
afs.tag_subject_of(('*', 'seda_appraisal_rule', '*'), 'inlined', 'hidden')
afs.tag_subject_of(('*', 'seda_appraisal_rule', '*'), 'main', 'inlined')
afs.tag_subject_of(('SEDAAppraisalRule', 'seda_appraisal_rule_code', '*'),
                   'main', 'inlined')
afs.tag_subject_of(('SEDAAppraisalRule', 'seda_appraisal_rule_duration', '*'),
                   'main', 'inlined')
afs.tag_subject_of(('SEDAAppraisalRuleCode', 'seda_appraisal_rule_code_value', '*'),
                   'main', 'attributes')
afs.tag_subject_of(('SEDAAppraisalRuleDuration', 'seda_appraisal_rule_duration_value', '*'),
                   'main', 'attributes')
affk.tag_subject_of(('SEDAAppraisalRuleDuration', 'seda_appraisal_rule_duration_value', '*'),
                    {'choices': rtype_concept_alt_label_choices})

# simple link-to-concept attributes (for archive, archive object and document)
for attr in ('seda_access_restriction_code',
             'seda_document_type_code', 'seda_file_type_code', 'seda_character_set_code'):
    afs.tag_subject_of(('*', attr, '*'), 'inlined', 'hidden')
    afs.tag_subject_of(('*', attr, '*'), 'main', 'inlined')
    afs.tag_subject_of(('*', attr + '_value', '*'), 'main', 'attributes')
    if attr == 'seda_access_restriction_code':  # Use a relation widget for access restriction code
        affk.tag_subject_of(('*', attr + '_value', '*'),
                            {'widget': rwdg.RelationFacetWidget(dialog_options={'width': 800})})
    elif attr == 'seda_document_type_code':  # Use a combo box for each other rtype
        affk.tag_subject_of(('*', attr + '_value', '*'),
                            {'choices': rtype_concept_pref_label_choices})
    else:
        affk.tag_subject_of(('*', attr + '_value', '*'),
                            {'choices': rtype_concept_alt_label_choices})

# children (for archive and archive object)
afs.tag_object_of(('*', 'seda_parent', '*'), 'main', 'hidden')
afs.tag_subject_of(('*', 'seda_parent', '*'), 'main', 'hidden')

# metadata attributes
aff.tag_attribute(('*', 'user_annotation'), SEDAMetaField)
afs.tag_attribute(('*', 'user_cardinality'), 'main', 'hidden')  # handled by SEDAMetaField

# clone relation
afs.tag_subject_of(('*', 'seda_clone_of', '*'), 'main', 'hidden')
afs.tag_object_of(('*', 'seda_clone_of', '*'), 'main', 'hidden')
afs.tag_subject_of(('*', 'seda_replace', '*'), 'main', 'hidden')
afs.tag_object_of(('*', 'seda_replace', '*'), 'main', 'hidden')


class SEDAMetaStringField(SEDAMetaFieldMixIn, ff.StringField):
    pass


aff.tag_attribute(('SEDAName', 'value'), SEDAMetaStringField)
afs.tag_attribute(('SEDAName', 'user_annotation'), 'main', 'hidden')
aff.tag_attribute(('SEDADescription', 'value'), SEDAMetaStringField)
afs.tag_attribute(('SEDADescription', 'user_annotation'), 'main', 'hidden')
aff.tag_attribute(('SEDAAppraisalRuleDuration', 'value'), SEDAMetaStringField)
afs.tag_attribute(('SEDAAppraisalRuleDuration', 'user_annotation'), 'main', 'hidden')


class SEDAMetaDateField(SEDAMetaFieldMixIn, ff.DateField):
    pass

aff.tag_attribute(('SEDADate', 'value'), SEDAMetaDateField)
afs.tag_attribute(('SEDADate', 'user_annotation'), 'main', 'hidden')


class SEDAMetaRelationField(SEDAMetaFieldMixIn, ff.RelationField):
    pass

aff.tag_attribute(('SEDADescriptionLevel', 'seda_description_level_value'),
                  SEDAMetaRelationField)
afs.tag_attribute(('SEDADescriptionLevel', 'user_annotation'), 'main', 'hidden')
aff.tag_attribute(('SEDAAppraisalRuleCode', 'seda_appraisal_rule_code_value'),
                  SEDAMetaRelationField)
afs.tag_attribute(('SEDAAppraisalRuleCode', 'user_annotation'), 'main', 'hidden')
aff.tag_attribute(('SEDAAccessRestrictionCode', 'seda_access_restriction_code_value'),
                  SEDAMetaRelationField)
afs.tag_attribute(('SEDAAccessRestrictionCode', 'user_annotation'), 'main', 'hidden')

aff.tag_attribute(('SEDAFileTypeCode', 'seda_file_type_code_value'), SEDAMetaRelationField)
afs.tag_attribute(('SEDAFileTypeCode', 'user_annotation'), 'main', 'hidden')
aff.tag_attribute(('SEDADocumentTypeCode', 'seda_document_type_code_value'), SEDAMetaRelationField)
afs.tag_attribute(('SEDADocumentTypeCode', 'user_annotation'), 'main', 'hidden')
aff.tag_attribute(('SEDACharacterSetCode', 'seda_character_set_code_value'), SEDAMetaRelationField)
afs.tag_attribute(('SEDACharacterSetCode', 'user_annotation'), 'main', 'hidden')

# hide parent objects
afs.tag_object_of(('*', 'seda_content_description', '*'), 'main', 'hidden')
afs.tag_object_of(('*', 'seda_appraisal_rule', '*'), 'main', 'hidden')
afs.tag_subject_of(('*', 'seda_keyword_of', '*'), 'main', 'hidden')
afs.tag_subject_of(('*', 'seda_parent', '*'), 'main', 'hidden')
pvs.tag_object_of(('*', 'seda_content_description', '*'), 'hidden')
pvs.tag_object_of(('*', 'seda_appraisal_rule', '*'), 'hidden')
pvs.tag_subject_of(('*', 'seda_keyword_of', '*'), 'hidden')
pvs.tag_subject_of(('*', 'seda_parent', '*'), 'hidden')
for rtype in ('seda_access_restriction_code', 'seda_document_type_code', 'seda_file_type_code',
              'seda_character_set_code'):
    afs.tag_object_of(('*', rtype, '*'), 'main', 'hidden')
    afs.tag_object_of(('*', rtype + '_value', '*'), 'main', 'hidden')
    pvs.tag_object_of(('*', rtype, '*'), 'hidden')
    pvs.tag_object_of(('*', rtype + '_value', '*'), 'hidden')

# finalize SEDAProfile form
affk.set_field_kwargs('SEDAProfile', 'ark', widget=fw.TextInput({'size': 80}), required=False)
affk.set_field_kwargs('SEDAProfile', 'title', widget=fw.TextInput({'size': 80}))


# shared configuration #############################################################################

affk.set_fields_order('SEDAContentDescription', ('user_cardinality', 'user_annotation',
                                                 'seda_description_level',
                                                 'seda_oldest_date', 'seda_latest_date',
                                                 'seda_keyword_of'))
pvds.set_fields_order('SEDAContentDescription', ('user_cardinality', 'user_annotation',
                                                 'seda_description_level', 'seda_description',
                                                 'seda_oldest_date', 'seda_latest_date',
                                                 'seda_keyword_of'))
affk.set_fields_order('SEDADescriptionLevel', ('seda_description_level_value', 'user_annotation'))
pvds.set_fields_order('SEDADescriptionLevel', ('seda_description_level_value', 'user_annotation'))
affk.set_fields_order('SEDAKeyword', ('user_cardinality', 'user_annotation',
                                      'seda_keyword_scheme', 'seda_keyword_value', ))
pvds.set_fields_order('SEDAKeyword', ('user_cardinality', 'user_annotation',
                                      'seda_keyword_scheme', 'seda_keyword_value', ))
affk.set_fields_order('SEDAAccessRestrictionCode', ('seda_access_restriction_code_value',
                                                    'user_cardinality', 'user_annotation'))
pvds.set_fields_order('SEDAAccessRestrictionCode', ('seda_access_restriction_code_value',
                                                    'user_cardinality', 'user_annotation'))
affk.set_fields_order('SEDAAppraisalRuleCode', ('seda_appraisal_rule_code_value',
                                                'user_annotation'))
pvds.set_fields_order('SEDAAppraisalRuleCode', ('seda_appraisal_rule_code_value',
                                                'user_annotation'))
affk.set_fields_order('ProfileArchiveObject', ('user_cardinality', 'user_annotation', 'seda_name'))
pvds.set_fields_order('ProfileArchiveObject', ('user_cardinality', 'user_annotation', 'seda_name'))

affk.set_fields_order('ProfileDocument', (
    'user_cardinality', 'user_annotation', 'seda_document_type_code', 'seda_description',
    'seda_character_set_code', 'seda_file_type_code'))
pvds.set_fields_order('ProfileDocument', (
    'user_cardinality', 'user_annotation', 'seda_document_type_code', 'seda_description',
    'seda_character_set_code', 'seda_file_type_code'))

affk.set_fields_order('SEDADocumentTypeCode',
                      ('seda_document_type_code_value', 'user_cardinality', 'user_annotation'))
pvds.set_fields_order('SEDADocumentTypeCode',
                      ('seda_document_type_code_value', 'user_cardinality', 'user_annotation'))
affk.set_fields_order('SEDAFileTypeCode',
                      ('seda_file_type_code_value', 'user_cardinality', 'user_annotation'))
pvds.set_fields_order('SEDAFileTypeCode',
                      ('seda_file_type_code_value', 'user_cardinality', 'user_annotation'))
affk.set_fields_order('SEDACharacterSetCode',
                      ('seda_character_set_code_value', 'user_cardinality', 'user_annotation'))
pvds.set_fields_order('SEDACharacterSetCode',
                      ('seda_character_set_code_value', 'user_cardinality', 'user_annotation'))


# views ############################################################################################

class NoLinkOneLineView(baseviews.OneLineView):
    __select__ = is_instance('SEDAName', 'SEDADescription', 'SEDADate', 'SEDAKeyword')

    def cell_call(self, row, col, **kwargs):
        """one line view without link to the entity """
        entity = self.cw_rset.get_entity(row, col)
        self.w(entity.dc_title())


class NoLinkInContextView(NoLinkOneLineView):
    __regid__ = 'incontext'


class NoLinkOutOfContextView(NoLinkOneLineView):
    __regid__ = 'outofcontext'


class ConceptAltLabelView(EntityView):
    __regid__ = 'saem.concept.alt'
    __select__ = is_instance('Concept')

    def entity_call(self, entity):
        desc = entity.preferred_label[0].label
        title = (entity.alternative_label[0].label
                 if entity.alternative_label else entity.dc_title())
        self.w(u'<a href="%s" title="%s">%s</a>' % (
            xml_escape(entity.absolute_url()), xml_escape(desc),
            xml_escape(title)))


# navigation #######################################################################################

class SEDAComponentsBreadcrumbsAdapter(ibreadcrumbs.IBreadCrumbsAdapter):
    """Breadcrumbs adapter pointing to /sedalib route when entity has no
    parent.
    """
    __select__ = (ibreadcrumbs.IBreadCrumbsAdapter.__select__
                  & relation_possible('seda_parent', role='subject')
                  & ~has_related_entities('seda_parent', role='subject'))

    def breadcrumbs(self, *args, **kwargs):
        return [(self._cw.build_url('sedalib'), self._cw._('SEDA components')),
                self.entity]


class SEDAComponentsBreadCrumbETypeVComponent(ibreadcrumbs.BreadCrumbEntityVComponent):
    """For proper display of the breadcrumb in the SEDA components list"""
    __select__ = (basecomponents.HeaderComponent.__select__
                  & multi_lines_rset() & is_instance('ProfileArchiveObject', 'ProfileDocument'))

    def render_breadcrumbs(self, w, contextentity, path):
        w(u'<a href="%s">%s</a>' % (self._cw.build_url('sedalib'),
                                    self._cw._('SEDA components')))


# hack autoform_section rtags using the draft 'selectable' rtags feature ###########################

def related_to_profile_archive(entity):
    """Return `1` if given `SEDAAccessRestrictionCode` entity is related through the
    `seda_access_restriction_code` relation to a `ProfileArchive`, else `0`.
    """
    if entity.has_eid():
        parent = entity.reverse_seda_access_restriction_code[0]
    else:
        # use .get since when related to a ProfileArchiveObject, the SEDAAccessRestrictionCode may
        # be acccessed through ajax hence missing the 'etype' key in form params
        # related_etype = entity._cw.form.get('etype')
        return 0  # XXX attempt to retrieve parent
    return int(parent.first_level)


paac_afs = selectable_uicfg(afs, __name__,
                            (is_instance('SEDAAccessRestrictionCode')
                             & score_entity(related_to_profile_archive)))
paac_afs.tag_attribute(('SEDAAccessRestrictionCode', 'user_cardinality'), 'main', 'hidden')


seda_clonable_afs = selectable_uicfg(afs, __name__,
                                     (relation_possible('seda_clone_of')
                                      & match_form_params(vid='copy')))

for rtype in ('seda_description',
              'seda_character_set_code',
              'seda_file_type_code'):
    seda_clonable_afs.tag_subject_of(('ProfileDocument', rtype, '*'),
                                     'main', 'hidden')
for rtype in ('seda_name',
              'seda_access_restriction_code',
              'seda_content_description',
              'seda_appraisal_rule'):
    seda_clonable_afs.tag_subject_of(('ProfileArchiveObject', rtype, '*'),
                                     'main', 'hidden')


# custom form renderers ############################################################################

class SEDAValueEntityInlinedFormRenderer(formrenderers.EntityInlinedFormRenderer):
    __select__ = formrenderers.EntityInlinedFormRenderer.__select__ & is_instance(
        'SEDAAppraisalRuleDuration', 'SEDAAppraisalRuleCode',
        'SEDAAccessRestrictionCode', 'SEDADescriptionLevel',
        'SEDAFileTypeCode', 'SEDACharacterSetCode', 'SEDADocumentTypeCode',
        'SEDAName', 'SEDADescription', 'SEDADate')

    def render_title(self, w, form, values):
        """Overriden from :class:`formrenderers.EntityInlinedFormRenderer` to skip the title"""
        if values['removejs']:
            w(u'<div class="iformTitle">')
            values['removemsg'] = self._cw.__('remove-inlined-entity-form')
            w(u'[<a href="javascript: %(removejs)s;$.noop();">%(removemsg)s</a>]'
              % values)
            w(u'</div>')

    def render_label(self, form, field):
        """Overriden from :class:`formrenderers.EntityInlinedFormRenderer` to set the label
        according to the relation to the parent entity.
        """
        if field.name.endswith('value'):
            # XXX rely on 'id' to search for the field holding the relation to the parent entity
            field = [f for f in form.fields if f.id and f.id.startswith('rel-')][0]
        return super(SEDAValueEntityInlinedFormRenderer, self).render_label(form, field)


class SEDACompoundEntityInlinedFormRenderer(formrenderers.EntityInlinedFormRenderer):
    __select__ = formrenderers.EntityInlinedFormRenderer.__select__ & is_instance(
        'SEDAContentDescription', 'SEDAAppraisalRule', 'SEDAKeyword')

    def open_form(self, w, form, values):
        """Overriden from :class:`formrenderers.EntityInlinedFormRenderer` to add the 'formGroup'
        class to then enclosing div. Beside this, HTML structure is left unchanged and so opened
        div are closed in :meth:`formrenderers.EntityInlinedFormRenderer.close_form`
        """
        try:
            w(u'<div id="div-%(divid)s" onclick="%(divonclick)s" class="formGroup">' % values)
        except KeyError:
            w(u'<div id="div-%(divid)s" class="formGroup">' % values)
        else:
            w(u'<div id="notice-%s" class="notice">%s</div>' % (
                values['divid'], self._cw.__('click on the box to cancel the deletion')))
        w(u'<div class="iformBody">')


# customization of relation widget for concepts ####################################################

class SearchForRelatedConceptsView(rwdg.SearchForRelatedEntitiesView):
    __abstract__ = True
    rtype = None
    has_creation_form = False

    def linkable_rset(self):
        """Return rset of entities to be displayed as possible values for the edited relation."""
        # XXX can't use alternative_label computed relation until
        # https://www.cubicweb.org/ticket/5497433 is fixed
        rql = ('Any C,CPL,CPLL,CAL,CALL,CD ORDERBY CPLL WHERE '
               'C in_scheme S, S scheme_relation RT, RT name %(rtype)s, '
               'C preferred_label CPL, CPL label CPLL, CAL? label_of C, CAL kind "alternative", '
               'CAL label CALL, C definition CD')
        return self._cw.execute(rql, {'rtype': self.rtype})

    @staticmethod
    def pref_label_label(concept):
        return concept.cw_rset.get_entity(concept.cw_row, 1).label

    @staticmethod
    def pref_label_column(w, concept):
        w(tags.a(SearchForRelatedConceptsView.pref_label_label(concept),
                 href=concept.absolute_url()))

    @staticmethod
    def alt_label(concept):
        if concept.cw_rset[concept.cw_row][3] is not None:
            return concept.cw_rset.get_entity(concept.cw_row, 3)
        return None


class SearchForAccessRestrictionConceptsView(SearchForRelatedConceptsView):
    rtype = 'seda_access_restriction_code'
    __select__ = SearchForRelatedConceptsView.__select__ & rwdg.edited_relation(rtype + '_value')


class SearchForDescriptionLevelConceptsView(SearchForRelatedConceptsView):
    rtype = 'seda_description_level'
    __select__ = SearchForRelatedConceptsView.__select__ & rwdg.edited_relation(rtype + '_value')


class SelectConceptEntitiesTableView(rwdg.SelectEntitiesTableView):
    """Table view of the selectable entities in the relation widget

    Selection of columns (and respective renderer) can be overridden by
    updating `columns` and `column_renderers` class attributes.
    """
    __select__ = (rwdg.SelectEntitiesTableView.__select__
                  & is_instance('Concept')
                  & rwdg.edited_relation([rtype + '_value' for rtype in (
                      'seda_access_restriction_code', 'seda_description_level')]))

    columns = rwdg.SelectEntitiesTableView.columns + ['alternative_label', 'definition']
    column_renderers = rwdg.SelectEntitiesTableView.column_renderers.copy()
    # speed-up things by considering rset shape from SearchForRelatedConceptsView
    column_renderers['entity'] = rwdg.SelectMainEntityColRenderer(
        sortfunc=SearchForRelatedConceptsView.pref_label_label,
        renderfunc=SearchForRelatedConceptsView.pref_label_column)
    column_renderers['alternative_label'] = tableview.RelatedEntityColRenderer(
        getrelated=SearchForRelatedConceptsView.alt_label)


# tree box #########################################################################################

def in_sedaprofile(entity):
    """Return 1 if entity is a (possibly indirect) child of a SEDAProfile."""
    while getattr(entity, 'seda_parent', None):
        entity = entity.seda_parent[0]
        if entity.cw_etype == 'SEDAProfile':
            return 1
    return 0


class SEDAProfileTreeBox(component.EntityCtxComponent):
    """display a box containing the whole SEDA profile tree"""
    __regid__ = 'saem.seda.tree'
    __select__ = one_line_rset() & (has_related_entities('seda_parent', role='subject')
                                    | has_related_entities('seda_parent', role='object'))
    title = _('SEDA profile objects tree')
    order = 10

    def render_body(self, w):
        self._cw.add_css('cubes.jqtree.css')
        self.entity.view('jqtree.treeview', w=w)


class ProfileArchiveObjectTreeBox(SEDAProfileTreeBox):
    __select__ = SEDAProfileTreeBox.__select__ & ~score_entity(in_sedaprofile)
    title = _('Profile archive object objects tree')


class SEDAIJQTreeAdapter(jqtree.IJQTreeAdapter):
    __select__ = jqtree.IJQTreeAdapter.__select__ & is_instance(
        'SEDAProfile', 'ProfileArchiveObject', 'ProfileDocument')

    def maybe_parent_of(self):
        parents = []
        rschema = self._cw.vreg.schema['seda_parent']
        for subjtype, objtype in rschema.rdefs:
            if objtype.type == self.entity.cw_etype:
                parents.append(subjtype.type)
        return parents

    def maybe_child(self):
        return 'seda_parent' in self.entity.e_schema.subjrels

    def reparent(self, peid):
        """Add a `seda_parent` relation between two entities."""
        rset = self._cw.execute(
            'SET C seda_parent P WHERE C eid %(c)s, P eid %(p)s',
            {'c': self.entity.eid, 'p': peid})
        return rset.rows


class SEDAJsonTreeView(jqtree.JsonTreeView):
    __select__ = jqtree.JsonTreeView.__select__ & is_instance(
        'SEDAProfile', 'ProfileArchiveObject', 'ProfileDocument')

    def entity_call(self, entity, **kwargs):
        root = entity.cw_adapt_to('ITree').root()
        if root.cw_etype == 'SEDAProfile' and not root.support_seda_exports:
            errors = defaultdict(list)
            for eid, msg in root.cw_adapt_to('SEDA-0.2.xsd').blockers():
                errors[eid].append(self._cw._(msg))
            self._cw.data['seda-export-errors'] = errors
        super(SEDAJsonTreeView, self).entity_call(entity, **kwargs)


class JQTreeErrorItemLabelView(jqtree.JQTreeItemLabelView):
    """Override default label view in case the entity has some seda export error."""
    __select__ = (jqtree.JQTreeItemLabelView.__select__
                  & score_entity(lambda x: x.eid in x._cw.data.get('seda-export-errors', ())))

    def entity_call(self, entity):
        super(JQTreeErrorItemLabelView, self).entity_call(entity)
        title = '\n'.join(self._cw.data['seda-export-errors'][entity.eid])
        self.w(tags.span(u'', title=title, klass='seda-export-error icon-attention'))


# life-cycle management ############################################################################

# Hide copy action for SEDA profiles
actions.CopyAction.__select__ = actions.CopyAction.__select__ & ~is_instance('SEDAProfile')


# Only show clone action for SEDA profiles if no clone has been created yet

class SEDAProfileCloneAction(CloneAction):
    __select__ = (CloneAction.__select__
                  & is_instance('SEDAProfile')
                  & ~has_related_entities('seda_replace', 'object')
                  & is_in_state('published'))
    title = _('new version')
    category = 'mainactions'

CloneAction.__select__ &= ~is_instance('SEDAProfile')


class NoWarningCopyFormView(editforms.CopyFormView):
    """display primary entity creation form initialized with values from another
    entity
    """
    __select__ = editforms.CopyFormView.__select__ & is_instance('SEDAProfile')

    def render_form(self, entity):
        """fetch and render the form"""
        # make a copy of entity to avoid altering the entity in the
        # request's cache.
        entity.complete()
        self.newentity = copy(entity)
        self.copying = entity
        self.newentity.eid = self._cw.varmaker.next()
        super(editforms.CopyFormView, self).render_form(self.newentity)
        del self.newentity


def workflow_state(entity):
    """Return the state of the given entity."""
    return entity.cw_adapt_to('IWorkflowable').state


class SEDAProfileRelatedVersionsComponent(component.EntityCtxComponent):
    """Output a box containing relevant SEDA Profiles related to the current one."""
    __regid__ = 'saem.seda.relatedprofiles'
    __select__ = (component.EntityCtxComponent.__select__ &
                  (has_related_entities('seda_replace', role='subject') |
                   has_related_entities('seda_replace', role='object')))
    context = 'incontext'
    order = 11
    title = _('Related versions')

    def predecessor(self):
        """Yield the profile of which the displayed profile is a new version (draft or published).
        """
        if (workflow_state(self.entity) in (u'published', u'draft') and
                self.entity.predecessor):
            yield self.entity.predecessor

    def current_version(self, state):
        """Yield the latest profile, either published or draft, that replaces the displayed one.

        ``state`` parameter must be either "draft" or "published".
        """
        assert state in (u'published', u'draft')
        successor = self.entity.successor
        while successor and workflow_state(successor) != state:
            successor = successor.successor
        if successor:
            yield successor

    def render_body(self, w):
        profile_state = workflow_state(self.entity)
        display_entities = []
        if profile_state == u'draft':
            display_entities += self.predecessor()
        elif profile_state == u'published':
            display_entities += self.current_version(u'draft')
            display_entities += self.predecessor()
        elif profile_state == u'deprecated':
            display_entities += self.current_version(u'published')
        if display_entities:
            w(u'<ul class="list-group">')
            for entity in display_entities:
                w(u'<li class="list-group-item">')
                w(tags.span(self._cw._(workflow_state(entity)), klass='badge'))
                entity.view('incontext', w=w)
                w(u'</li>')
            w(u'</ul>')


# SEDA import ######################################################################################

def _import_div_id(entity):
    return 'importDiv%s' % entity.eid


class SEDAPrimaryView(primary.PrimaryView):
    """Overriden to add an extra empty div that may be used to insert import's relation widget"""
    __select__ = (primary.PrimaryView.__select__
                  & is_instance('SEDAProfile', 'ProfileArchiveObject')
                  & has_permission('update'))

    def entity_call(self, entity, **kwargs):
        super(SEDAPrimaryView, self).entity_call(entity, **kwargs)
        rwdg.boostrap_dialog(self.w, self._cw._, _import_div_id(entity), u'')


class SEDAImportAction(action.Action):
    __abstract__ = True

    __select__ = (action.Action.__select__
                  & one_line_rset()
                  & has_permission('update'))

    category = 'moreactions'
    submenu = _('import menu')
    etype = None  # to be specified in concrete classes

    @property
    def title(self):
        return self._cw._(self.etype).lower()

    def fill_menu(self, box, menu):
        # when there is only one item in the sub-menu, replace the sub-menu by
        # item's title prefixed by 'add'
        menu.label_prefix = self._cw._('import menu')
        super(SEDAImportAction, self).fill_menu(box, menu)

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        root = entity.cw_adapt_to('ITree').root()
        relation = 'seda_clone_of:%s:subject' % self.etype
        search_url = self._cw.build_url('ajax', fname='view', vid='search_related_entities',
                                        __modal=1, multiple='1', relation=relation,
                                        etype=root.cw_etype, target=root.eid)
        title = self._cw._('Search entity to import')
        return configure_relation_widget(self._cw, _import_div_id(entity), search_url, title,
                                         True, uilib.js.saem.buildSedaImportValidate(entity.eid))


class SEDAImportProfileArchiveObjectAction(SEDAImportAction):
    __regid__ = 'saem_ref.seda.import.documentunit'
    __select__ = SEDAImportAction.__select__ & is_instance('SEDAProfile', 'ProfileArchiveObject')
    etype = 'ProfileArchiveObject'


class SEDAImportProfileDocumentAction(SEDAImportAction):
    __regid__ = 'saem_ref.seda.import.attachment'
    __select__ = SEDAImportAction.__select__ & is_instance('ProfileArchiveObject')
    etype = 'ProfileDocument'


class SEDADoImportView(controller.Controller):
    __regid__ = 'saem_ref.seda.doimport'
    __select__ = match_form_params('eid', 'cloned')

    def publish(self, rset=None):
        parent = self._cw.entity_from_eid(int(self._cw.form['eid']))
        original = self._cw.entity_from_eid(int(self._cw.form['cloned']))
        clone = copy_entity(original, seda_parent=parent, seda_clone_of=original)
        msg = self._cw._('{0} has been imported').format(clone.dc_title())
        raise Redirect(parent.absolute_url(__message=msg))


class SearchForDocumentUnitToImportView(rwdg.SearchForRelatedEntitiesView):
    __select__ = (SearchForRelatedConceptsView.__select__
                  & rwdg.edited_relation('seda_clone_of')
                  & match_form_params('target'))
    title = None

    def linkable_rset(self):
        """Return rset of entities to be displayed as possible values for the edited relation."""
        tetype = self._cw.form['relation'].split(':')[1]
        target = int(self._cw.form['target'])
        rql = ('Any X,MD ORDERBY MD DESC WHERE X is %s, X modification_date MD, '
               'NOT X seda_parent P, NOT X eid %%(target)s') % tetype
        return self._cw.execute(rql, {'target': target})


class SelectProfileDocumentEntitiesTableView(rwdg.SelectEntitiesTableView):
    """Table view for ProfileDocument entities in a relation widget."""

    __select__ = (rwdg.SelectEntitiesTableView.__select__
                  & is_instance('ProfileDocument')
                  & rwdg.edited_relation('seda_clone_of'))

    columns = rwdg.SelectEntitiesTableView.columns + ['seda_description', 'seda_file_type_code']
    column_renderers = rwdg.SelectEntitiesTableView.column_renderers.copy()
    column_renderers['seda_description'] = tableview.RelatedEntityColRenderer(
        getrelated=lambda x: x.description)
    column_renderers['seda_file_type_code'] = tableview.RelatedEntityColRenderer(
        getrelated=lambda x: x.file_type_code, vid='saem.seda.concept-value')


class SelectProfileArchiveObjectEntitiesTableView(rwdg.SelectEntitiesTableView):
    """Table view ProfileArchiveObject entities in the relation widget."""

    __select__ = (rwdg.SelectEntitiesTableView.__select__
                  & is_instance('ProfileArchiveObject')
                  & rwdg.edited_relation('seda_clone_of'))

    columns = rwdg.SelectEntitiesTableView.columns[:]
    columns += ['seda_content_description', 'seda_description_level',
                'seda_appraisal_rule']
    column_renderers = rwdg.SelectEntitiesTableView.column_renderers.copy()
    column_renderers['entity'] = rwdg.SelectMainEntityColRenderer(vid='saem.seda.du.name')
    column_renderers['seda_name'] = tableview.RelatedEntityColRenderer(
        getrelated=lambda x: x.seda_name[0])
    column_renderers['seda_content_description'] = tableview.RelatedEntityColRenderer(
        getrelated=lambda x: x.content_description and x.content_description.description)
    column_renderers['seda_description_level'] = tableview.RelatedEntityColRenderer(
        getrelated=lambda x: x.content_description and x.content_description.description_level,
        vid='saem.seda.concept-value')
    column_renderers['seda_appraisal_rule'] = tableview.RelatedEntityColRenderer(
        getrelated=lambda x: x.appraisal_rule and x.appraisal_rule.appraisal_rule_code,
        vid='saem.seda.concept-value')


class DocumentUnitNameView(EntityView):
    __regid__ = 'saem.seda.du.name'
    __select__ = is_instance('ProfileArchiveObject')

    def entity_call(self, entity):
        self.w(tags.a(entity.name.dc_title(), href=entity.absolute_url(), target='_blank'))


class ConceptValueView(EntityView):
    """Display related concept if any, else attempt to fallback to user annotation."""
    __abstract__ = True
    __regid__ = 'saem.seda.concept-value'
    concept_rtype = None
    label_rtype = 'preferred_label'

    def label(self, entity):
        concepts = getattr(entity, self.concept_rtype)
        if concepts:
            return getattr(concepts[0], self.label_rtype)[0].printable_value('label')
        return None

    def entity_call(self, entity):
        label = self.label(entity)
        if label is None:
            label = entity.printable_value('user_annotation')
        if label:
            self.w(label)


class SEDAAppraisalRuleCodeConceptValueView(ConceptValueView):
    __select__ = is_instance('SEDAAppraisalRuleCode')
    concept_rtype = 'seda_appraisal_rule_code_value'


class SEDADescriptionLevelConceptValueView(ConceptValueView):
    __select__ = is_instance('SEDADescriptionLevel')
    concept_rtype = 'seda_description_level_value'


class SEDAFileTypeCodeConceptValueView(ConceptValueView):
    __select__ = is_instance('SEDAFileTypeCode')
    concept_rtype = 'seda_file_type_code_value'
    label_rtype = 'alternative_label'


# SEDA download ####################################################################################

class SEDA1DownloadAction(action.Action):
    __regid__ = 'saem_ref.seda1'
    __select__ = (action.Action.__select__
                  & one_line_rset()
                  & is_instance('SEDAProfile')
                  & score_entity(lambda x: 'SEDA-1.0.xsd' in x.support_seda_exports))

    title = _('SEDA 1.0 export')
    category = 'moreactions'
    seda_download_vid = 'saem_ref.seda1'

    def url(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        return entity.absolute_url(vid=self.seda_download_vid)


class SEDA1DownloadView(idownloadable.DownloadView):
    """SEDA 1.0 download view"""
    __regid__ = 'saem_ref.seda1'
    __select__ = one_line_rset() & is_instance('SEDAProfile')

    http_cache_manager = httpcache.NoHTTPCacheManager
    seda_adapter_id = 'SEDA-1.0.xsd'

    def set_request_content_type(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.seda_adapter_id)
        self._cw.set_content_type(adapter.content_type, filename=adapter.file_name,
                                  encoding=adapter.encoding, disposition='attachment')

    def call(self):
        entity = self.cw_rset.get_entity(self.cw_row or 0, self.cw_col or 0)
        adapter = entity.cw_adapt_to(self.seda_adapter_id)
        self.w(adapter.dump())


class SEDA02DownloadAction(SEDA1DownloadAction):
    __regid__ = 'saem_ref.seda02'
    __select__ = (action.Action.__select__
                  & one_line_rset()
                  & is_instance('SEDAProfile')
                  & score_entity(lambda x: 'SEDA-0.2.xsd' in x.support_seda_exports))
    title = _('SEDA 0.2 export')
    seda_download_vid = 'saem_ref.seda02'


class SEDA02DownloadView(SEDA1DownloadView):
    """SEDA 0.2 download view"""
    __regid__ = 'saem_ref.seda02'
    seda_adapter_id = 'SEDA-0.2.xsd'


# SEDA lib components ##############################################################################


class AddEntityComponent(component.CtxComponent):
    """Component with 'add' link to be displayed in 'same etype' views usually 'SameETypeListView'.
    """
    __regid__ = 'saem_ref.add_entity'
    __select__ = (component.CtxComponent.__select__
                  & multi_lines_rset()
                  & has_permission('add')
                  & is_instance('ProfileArchiveObject', 'ProfileDocument'))
    context = 'navtop'

    def render_body(self, w):
        req = self._cw
        links = [add_etype_link(req, etype, req._(etype), klass='')
                 for etype in ('ProfileArchiveObject', 'ProfileDocument')]
        w(index_dropdown_button(req._('Add a new component'), links))


class SEDALibView(baseviews.SameETypeListView):
    __regid__ = 'saem_ref.sedalib'
    __select__ = (multi_lines_rset()
                  & is_instance('ProfileArchiveObject', 'ProfileDocument'))

    @property
    def title(self):
        return self._cw._('SEDA components')
